from typing import Tuple, List, Optional, Union

from citrine.resources.project import Project
from citrine.informatics.predictors import (
    MolecularStructureFeaturizer,
    ChemicalFormulaFeaturizer,
    MeanPropertyPredictor,
    AutoMLPredictor,
    GraphPredictor
)
from citrine.informatics.descriptors import (
    Descriptor,
    FormulationDescriptor,
    RealDescriptor,
    CategoricalDescriptor,
    ChemicalFormulaDescriptor,
    MolecularStructureDescriptor,
)
from citrine.informatics.data_sources import DataSource


def build_mean_feature_property_predictors(
        *,
        project: Project,
        featurizer: Union[MolecularStructureFeaturizer, ChemicalFormulaFeaturizer],
        formulation_descriptor: FormulationDescriptor,
        p: int,
        impute_properties: bool = True,
        make_all_ingredients_model: bool = True,
        labels: Optional[List[str]] = None
) -> Tuple[List[MeanPropertyPredictor], List[Descriptor]]:
    """[ALPHA] Combine a featurizer model with mean property models.

    Given a featurizer, produce "mean property" models that calculate the mean of the
    properties calculated by the featurizer. This is useful if you do not directly know
    the numerical properties of ingredients in a formulation, but instead know, for example,
    the molecular structure. This builder method determines the real-valued properties that the
    featurizer calculates and builds mean property models that use them as input properties.

    Parameters
    ----------
    project: Project
        Project that contains the predictor
    featurizer: Union[MolecularStructureFeaturizer, ChemicalFormulaFeaturizer]
        A model that is being used to featurize formulation ingredients. Currently only
        accepts a molecular structure featurizer or a chemical formula featurizer.
    formulation_descriptor: FormulationDescriptor
        Descriptor that represents the formulation being featurized.
    p: int
        Power of the generalized mean. Only integer powers are supported.
    impute_properties: bool
        Whether to impute missing ingredient properties by averaging over the entire dataset.
        If ``False`` all ingredients must define values for all featurized properties.
        Otherwise, the row will not be featurized.
    make_all_ingredients_model: bool
        Whether to create a mean property predictor that calculates the mean over all ingredients.
        If False, models are only constructed for specified labels. Must be True if no labels
        are specified.
    labels: Optional[List[str]]
        List of labels for which a mean property predictor should be created.

    Returns
    -------
    Tuple[List[MeanPropertyPredictor], List[Descriptor]]
        List of mean property predictors that should be incorporated into the graph predictor,
        and a list of all the output descriptors produced by these models. There will be one
        model for each label specified, and one model for all ingredients if ``all_ingredients``
        is set to ``True``. In the common case, the output descriptors will all be used
        as inputs to one or more ML models.

    """
    if isinstance(featurizer, (MolecularStructureFeaturizer, ChemicalFormulaFeaturizer)):
        input_descriptor = featurizer.input_descriptor
    else:
        raise TypeError(f"Featurizer of type {type(featurizer)} is not supported.")

    if labels is None:
        labels = []
    if not isinstance(labels, (list, set)):
        raise TypeError("labels must be specified as a list or set of strings.")
    if make_all_ingredients_model:
        labels = [None] + labels
    if len(labels) == 0:
        msg = "No mean property predictors requested. " \
              "Set make_all_ingredients_model to True and/or specify labels."
        raise ValueError(msg)

    properties = project.descriptors.from_predictor_responses(
        predictor=featurizer, inputs=[input_descriptor]
    )
    real_properties = [desc for desc in properties if isinstance(desc, RealDescriptor)]
    if len(real_properties) == 0:
        msg = "Featurizer did not return any real properties to calculate the means of."
        raise RuntimeError(msg)

    predictors = [
        _build_mean_property_predictor(
            ingredient_descriptor=input_descriptor,
            formulation_descriptor=formulation_descriptor,
            properties=real_properties,
            p=p,
            impute_properties=impute_properties,
            label=label
        )
        for label in labels
    ]

    all_outputs = [
        output
        for single_model_outputs in [
            project.descriptors.from_predictor_responses(
                predictor=predictor, inputs=[formulation_descriptor]
            )
            for predictor in predictors
        ]
        for output in single_model_outputs
    ]

    return predictors, all_outputs


def _build_mean_property_predictor(
        ingredient_descriptor: Descriptor,
        formulation_descriptor: FormulationDescriptor,
        properties: List[RealDescriptor],
        p: int,
        impute_properties: bool,
        label: Optional[str]
) -> MeanPropertyPredictor:
    """Build a MeanPropertyPredictor for given specifications."""
    name = f"mean of {ingredient_descriptor.key} features"
    if p != 1:
        name = f"{p}-{name}"
    if label is not None:
        name += f" for label {label}"
    name += f" in {formulation_descriptor.key}"

    return MeanPropertyPredictor(
        name=name,
        description="",
        input_descriptor=formulation_descriptor,
        properties=properties,
        p=p,
        impute_properties=impute_properties,
        label=label
    )


# Add simple ml builder
def build_simple_ml(
    *,
    project: Project,
    name: str,
    description: str,
    inputs: List[Descriptor],
    outputs: List[Descriptor],
    latent_variables: List[Descriptor],
    training_data: Optional[List[DataSource]] = None
) -> GraphPredictor:
    """
    Build a graph predictor that connects all inputs to all latent variables and outputs.

    The structure of the returned, unregistered graph predictor will be analogous to the
    structure returned by SimpleMLPredictor.

    Parameters
    ----------
    project: Project
        Project that contains the predictor to be built
    name: str
        name of the returned predictor
    description: str
        the description of the predictor
    inputs: list[Descriptor]
        Descriptors that represent inputs to relations
    outputs: list[Descriptor]
        Descriptors that represent outputs of relations
    latent_variables: list[Descriptor]
        Descriptors that are predicted from inputs and used when predicting the outputs
    training_data: Optional[List[DataSource]]
        Sources of training data. Each can be either a CSV or an GEM Table. Candidates from
        multiple data sources will be combined into a flattened list and de-duplicated by uid and
        identifiers. de-duplication is performed if a uid or identifier is shared between two or
        more rows. The content of a de-duplicated row will contain the union of data across all
        rows that share the same uid or at least 1 identifier. Training data is unnecessary if the
        predictor is part of a graph that includes all training data required by this predictor.

    Returns
    -------
    GraphPredictor
        GraphPredictor connecting all inputs to all latent variables and all inputs and latent
        variables to all outputs.

    """
    ml_model_feature_descriptors = [d for d in inputs
                                    if isinstance(d, (RealDescriptor, CategoricalDescriptor))]
    chemical_formula_descriptors = [d for d in inputs
                                    if isinstance(d, ChemicalFormulaDescriptor)]
    molecular_structure_descriptors = [d for d in inputs
                                       if isinstance(d, MolecularStructureDescriptor)]

    chemical_formula_featurizers = [
        ChemicalFormulaFeaturizer(
            name=f'{chemical_formula_descriptor.key} featurizer',
            description='',
            input_descriptor=chemical_formula_descriptor,
            features=['standard'],
        ) for chemical_formula_descriptor in chemical_formula_descriptors
    ]

    molecular_structure_featurizers = [
        MolecularStructureFeaturizer(
            name=f'{molecular_structure_descriptor.key} featurizer',
            description='',
            input_descriptor=molecular_structure_descriptor,
            features=['standard'],
        ) for molecular_structure_descriptor in molecular_structure_descriptors
    ]

    # TODO: Fix needing to pass the project (if possible)
    featurizer_responses = []
    for featurizer in chemical_formula_featurizers + molecular_structure_featurizers:
        featurizer_responses.extend(
            project.descriptors.from_predictor_responses(
                predictor=featurizer,
                inputs=[featurizer.input_descriptor],
            )
        )

    automl_lv_predictors = [
        AutoMLPredictor(
            name=f'{latent_variable_descriptor.key} Predictor',
            description='',
            inputs=ml_model_feature_descriptors + featurizer_responses,
            outputs=[latent_variable_descriptor],
        ) for latent_variable_descriptor in latent_variables
    ]

    automl_output_predictors = [
        AutoMLPredictor(
            name=f'{output_descriptor.key} Predictor',
            description='',
            inputs=ml_model_feature_descriptors + latent_variables + featurizer_responses,
            outputs=[output_descriptor],
        ) for output_descriptor in outputs
    ]

    predictor = GraphPredictor(
        name=name,
        description=description,
        predictors=[
            * chemical_formula_featurizers,
            * molecular_structure_featurizers,
            * automl_lv_predictors,
            * automl_output_predictors
        ],
        training_data=training_data
    )

    return predictor



"""Resources that represent process run data objects."""
from typing import List, Dict, Optional, Type, Union, Iterator
from uuid import UUID

from citrine._rest.resource import GEMDResource
from citrine._serialization.properties import List as PropertyList
from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import String, Object, LinkOrElse
from citrine.resources.object_runs import ObjectRun, ObjectRunCollection
from gemd.entity.attribute.condition import Condition
from gemd.entity.attribute.parameter import Parameter
from gemd.entity.file_link import FileLink
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.object.process_run import ProcessRun as GEMDProcessRun
from gemd.entity.object.process_spec import ProcessSpec as GEMDProcessSpec
from gemd.entity.source.performed_source import PerformedSource


class ProcessRun(GEMDResource['ProcessRun'], ObjectRun, GEMDProcessRun, typ=GEMDProcessRun.typ):
    """
    A process run.

    Processes transform zero or more input materials into exactly one output material.

    Parameters
    ----------
    name: str
        Name of the process run.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/gemd-docs/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/gemd-docs/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the process run.
    conditions: List[Condition], optional
        Conditions under which this process run occurs.
    parameters: List[Parameter], optional
        Parameters of this process run.
    spec: ProcessSpec
        Spec for this process run.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.
    source: PerformedSource, optional
        Information about the person who performed the run and when.

    Attributes
    ----------
    output_material: MaterialRun
        The material run that this process run produces. The link is established by creating
        the material run and settings its `process` field to this process run.

    ingredients: List[IngredientRun]
        Ingredient runs that act as inputs to this process run. The link is established by
        creating each ingredient run and setting its `process` field to this process run.

    """

    _response_key = GEMDProcessRun.typ  # 'process_run'

    name = String('name', override=True, use_init=True)
    conditions = PropertyOptional(PropertyList(Object(Condition)), 'conditions', override=True)
    parameters = PropertyOptional(PropertyList(Object(Parameter)), 'parameters', override=True)
    spec = PropertyOptional(LinkOrElse(GEMDProcessSpec), 'spec', override=True, use_init=True,)
    source = PropertyOptional(Object(PerformedSource), "source", override=True)

    def __init__(self,
                 name: str,
                 *,
                 uids: Optional[Dict[str, str]] = None,
                 tags: Optional[List[str]] = None,
                 notes: Optional[str] = None,
                 conditions: Optional[List[Condition]] = None,
                 parameters: Optional[List[Parameter]] = None,
                 spec: Optional[GEMDProcessSpec] = None,
                 file_links: Optional[List[FileLink]] = None,
                 source: Optional[PerformedSource] = None):
        if uids is None:
            uids = dict()
        super(ObjectRun, self).__init__()
        GEMDProcessRun.__init__(self, name=name, uids=uids,
                                tags=tags, conditions=conditions, parameters=parameters,
                                spec=spec, file_links=file_links, notes=notes, source=source)

    def __str__(self):
        return '<Process run {!r}>'.format(self.name)


class ProcessRunCollection(ObjectRunCollection[ProcessRun]):
    """Represents the collection of all process runs associated with a dataset."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/process-runs'
    _dataset_agnostic_path_template = 'projects/{project_id}/process-runs'
    _individual_key = 'process_run'
    _collection_key = 'process_runs'
    _resource = ProcessRun

    @classmethod
    def get_type(cls) -> Type[ProcessRun]:
        """Return the resource type in the collection."""
        return ProcessRun

    def list_by_spec(self,
                     uid: Union[UUID, str, LinkByUID, GEMDProcessSpec]
                     ) -> Iterator[ProcessRun]:
        """
        Get the process runs using the specified process spec.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDProcessSpec]
            A representation of the process spec whose process run usages are to be located.

        Returns
        -------
        Iterator[ProcessRun]
            The process runs using the specified process spec.

        """
        return self._get_relation('process-specs', uid=uid)



import csv
from io import StringIO
from typing import List, Iterable

from citrine.resources.project import Project
from citrine.resources.gemtables import GemTable

from citrine.informatics.objectives import Objective, ScalarMinObjective
from citrine.informatics.scores import Score, LIScore


def create_default_score(
    *, objectives: List[Objective], project: Project, table: GemTable
) -> Score:
    """[ALPHA] Create a default score from the provided objectives.

    Reads the provided table from S3 storage,
    and parses the columns corresponding to each objective's descriptor key
    to extract relevant baseline values.

    The baselines are determined as the min/max of the column data,
    depending on whether the objective is a ScalarMinObjective or ScalarMaxObjective.

    Parameters
    ----------
    objectives: List[Objective]
        List of objectives to include in a score.
    project: Project
        Project to use when accessing the Citrine Platform.
    table: GemTable
        Table to read baseline data from.
        Must contain column definitions matching the descriptor keys
        found in each objective.

    Returns
    ----------
    Score
        The default constructed score.
        Currently, this is a LIScore containing containing all provided objectives.

    """
    table_data = project.tables.read_to_memory(table)
    data_io = StringIO(table_data)
    reader = csv.DictReader(data_io, delimiter=",")

    objectives = objectives if isinstance(objectives, Iterable) else [objectives]

    header_map = {}
    objective_values = {}
    for objective in objectives:
        key = objective.descriptor_key
        objective_values[key] = []

        # Descriptor key has to be mapped to fieldnames of table data
        # Search for mean column, avoid matching on std columns
        full_key = next(filter(lambda col: f"{key}~Mean" in col, reader.fieldnames), None)
        if full_key is None:
            raise ValueError(f"Key '{key}' not found in GEM table headers.")
        header_map[key] = full_key

    # Iterate table data to extract baseline info
    for row in reader:
        for objective in objectives:
            key = objective.descriptor_key
            full_key = header_map[key]
            try:
                objective_values[key].append(float(row[full_key]))
            except ValueError:
                continue

    # Determine max/min of extracted values for baseline
    baselines = []
    for objective in objectives:
        key = objective.descriptor_key
        if len(objective_values[key]) == 0:
            raise ValueError(f"Unable to parse values for '{key}' as numeric.")

        comparator = min if isinstance(objective, ScalarMinObjective) else max
        baselines.append(comparator(objective_values[key]))

    score = LIScore(objectives=objectives, baselines=baselines)
    return score



import json
from typing import List, Union, Tuple, Optional
from uuid import UUID

from gemd.entity.base_entity import BaseEntity
from gemd.entity.link_by_uid import LinkByUID

from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.resources.api_error import ApiError
from citrine.jobs.job import _poll_for_job_completion
from citrine.resources.data_concepts import _make_link_by_uid


def _async_gemd_batch_delete(
        id_list: List[Union[LinkByUID, UUID, str, BaseEntity]],
        project_id: UUID,
        session: Session,
        dataset_id: Optional[UUID],
        timeout: float = 2 * 60,
        polling_delay: float = 1.0
) -> List[Tuple[LinkByUID, ApiError]]:
    """
    Shared implementation of Async GEMD Batch deletion.

    See documentation for _gemd_batch_delete. The only difference is that this version polls for
    an asynchronous result and can tolerate a very long runtime that the synchronous version
    cannot. Because this version can tolerate a long runtime, this versions allows for the
    removal of attribute templates.

    Parameters
    ----------
    id_list: List[Union[LinkByUID, UUID, str, BaseEntity]]
        A list of the IDs of data objects to be removed. They can be passed
        as a LinkByUID tuple, a UUID, a string, or the object itself. A UUID
        or string is assumed to be a Citrine ID, whereas a LinkByUID or
        BaseEntity can also be used to provide an external ID.

    project_id: UUID
        The Project ID to use in the delete request.

    session: Session
        The Citrine session.

    dataset_id: Optional[UUID] = None
        An optional dataset ID, which if provided will mandate that all GEMD objects
        must be within the given dataset.

    timeout: float
        Amount of time to wait on the job (in seconds) before giving up. Defaults
        to 2 minutes. Note that this number has no effect on the underlying job
        itself, which can also time out server-side.

    polling_delay: float
        How long to delay between each polling retry attempt.

    Returns
    -------
    List[Tuple[LinkByUID, ApiError]]
        A list of (LinkByUID, api_error) for each failure to delete an object.
        Note that this method doesn't raise an exception if an object fails to be
        deleted.

    """
    scoped_uids = []
    for uid in id_list:  # And now normalize to id/scope pairs
        link_by_uid = _make_link_by_uid(uid)
        scoped_uids.append({'scope': link_by_uid.scope, 'id': link_by_uid.id})

    body = {'ids': scoped_uids}

    if dataset_id is not None:
        body.update({'dataset_id': str(dataset_id)})

    path = format_escaped_url('/projects/{project_id}/gemd/async-batch-delete',
                              project_id=project_id
                              )
    response = session.post_resource(path, body)

    job_id = response["job_id"]

    return _poll_for_async_batch_delete_result(project_id, session, job_id, timeout, polling_delay)


def _poll_for_async_batch_delete_result(
        project_id: UUID,
        session: Session,
        job_id: str,
        timeout: float,
        polling_delay: float
) -> List[Tuple[LinkByUID, ApiError]]:
    """
    Poll for the result of an asynchronous batch delete (or a deletion of dataset contents).

    Parameters
    ----------
    project_id: UUID
        The Project ID to use in the delete request.

    session: Session
        The Citrine session.

    job_id: str
        The asynchronous Job ID.

    timeout: float
        Amount of time to wait on the job (in seconds) before giving up.
        Note that this number has no effect on the underlying job itself,
        which can also time out server-side.

    polling_delay: float
        How long to delay between each polling retry attempt.

    Returns
    -------
    List[Tuple[LinkByUID, ApiError]]
        A list of (LinkByUID, api_error) for each failure to delete an object.
        Note that this method doesn't raise an exception if an object fails to be
        deleted.

    """
    response = _poll_for_job_completion(session, project_id, job_id, timeout=timeout,
                                        polling_delay=polling_delay)

    return [(LinkByUID(f['id']['scope'], f['id']['id']), ApiError.build(f['cause']))
            for f in json.loads(response.output.get('failures', '[]'))]



from typing import Type, Optional

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable


__all__ = [
    'DesignCandidate',
    'HierarchicalDesignCandidate',
    'DesignMaterial',
    'HierarchicalDesignMaterial',
    'SampleSearchSpaceResultCandidate',
    'DesignVariable',
    'MeanAndStd',
    'TopCategories',
    'Mixture',
    'ChemicalFormula',
    'MolecularStructure',
]


class DesignVariable(PolymorphicSerializable["DesignVariable"]):
    """Classes containing data corresponding to individual descriptors.

    If you think of materials as being represented as a set of (descriptor, value) pairs,
    these are simplified representations of the values.
    """

    def __init__(self, arg):
        pass  # pragma: no cover

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        return {
            "R": MeanAndStd,
            "C": TopCategories,
            "M": Mixture,
            "F": ChemicalFormula,
            "S": MolecularStructure
        }[data["type"]]


class MeanAndStd(Serializable["MeanAndStd"], DesignVariable):
    """The mean and standard deviation of a continuous distribution.

    This does not imply that the distribution is Normal.
    """

    mean = properties.Float('m')
    """:float: mean of the continuous distribution"""
    std = properties.Float('s')
    """:float: standard deviation of the continuous distribution"""
    typ = properties.String('type', default='R', deserializable=False)
    """:str: polymorphic type code"""

    def __init__(self, *, mean: float, std: float):
        self.mean = mean
        self.std = std
        pass  # pragma: no cover


class TopCategories(Serializable["CategoriesAndProbabilities"], DesignVariable):
    """The category names and probabilities for the most probable categories.

    This list is truncated: these are the most probable categories but other categories
    may have non-zero probabilities.
    """

    probabilities = properties.Mapping(properties.String, properties.Float, 'cp')
    """:Dict[str, float]: mapping from category names to their probabilities"""
    typ = properties.String('type', default='C', deserializable=False)
    """:str: polymorphic type code"""

    def __init__(self, *, probabilities: dict):
        self.probabilities = probabilities
        pass  # pragma: no cover


class Mixture(Serializable["Mixture"], DesignVariable):
    """Most likely quantity values for all of the components in a mixture.

    This is a complete list of components with non-zero quantities; there is no
    truncation (but there may be rounding).
    """

    quantities = properties.Mapping(properties.String, properties.Float, 'q')
    """:Dict[str, float]: mapping from ingredient identifiers to their quantities"""
    labels = properties.Mapping(properties.String, properties.Set(properties.String), 'l')
    """:Dict[str, Set[str]]: mapping from label identifiers to their associated ingredients"""
    typ = properties.String('type', default='M', deserializable=False)
    """:str: polymorphic type code"""

    def __init__(self, *, quantities: dict, labels: Optional[dict] = None):
        self.quantities = quantities
        self.labels = labels or {}
        pass  # pragma: no cover


class ChemicalFormula(Serializable["ChemicalFormula"], DesignVariable):
    """Chemical formula as a string."""

    formula = properties.String('f')
    """:str: chemical formula"""
    typ = properties.String('type', default='F', deserializable=False)
    """:str: polymorphic type code"""

    def __init__(self, *, formula: str):
        self.formula = formula
        pass  # pragma: no cover


class MolecularStructure(Serializable["MolecularStructure"], DesignVariable):
    """SMILES string representation of a molecular structure."""

    smiles = properties.String('s')
    """:str: SMILES string"""
    typ = properties.String('type', default='S', deserializable=False)
    """:str: polymorphic type code"""

    def __init__(self, *, smiles: str):
        self.smiles = smiles
        pass  # pragma: no cover


class DesignMaterial(Serializable["DesignMaterial"]):
    """Description of the material that was designed, as a set of DesignVariables."""

    material_id = properties.UUID('identifiers.id')
    """:UUID: unique internal Citrine id of the material"""
    identifiers = properties.List(properties.String, 'identifiers.external', default=[])
    """:List[str]: globally unique identifiers assigned to the material"""
    process_template = properties.Optional(properties.UUID, 'identifiers.process_template')
    """:Optional[UUID]: GEMD process template that describes the process to create this material"""
    material_template = properties.Optional(properties.UUID, 'identifiers.material_template')
    """:Optional[UUID]: GEMD material template that describes this material"""
    values = properties.Mapping(properties.String, properties.Object(DesignVariable), 'vars')
    """:Dict[str, DesignVariable]: mapping from descriptor keys to the value for this material"""

    def __init__(self, *, values: dict):
        self.values = values
        pass


class HierarchicalDesignMaterial(Serializable["HierarchicalDesignMaterial"]):
    """Description of a designed material as a set of connected individual materials.

    A hierarchical material contains a root material, containing identifiers and variables,
    along any of sub-materials appearing in the history of the root.
    Connections between the root and sub-materials are contained in the mixtures dictionary
    that associates each material (by Citrine ID) with the ingredients that comprise it.
    """

    root = properties.Object(DesignMaterial, 'terminal')
    """:DesignMaterial: root material containing features and predicted properties"""
    sub_materials = properties.List(properties.Object(DesignMaterial), 'sub_materials')
    """:List[DesignMaterial]: all other materials appearing in the history of the root"""
    mixtures = properties.Mapping(properties.UUID, properties.Object(Mixture), 'mixtures')
    """:Dict[UUID, Mixture]: mapping from Citrine ID to components the material is composed of"""


class DesignCandidate(Serializable["DesignCandidate"]):
    """A candidate material generated by the Citrine Platform.

    This class represents the candidate computed by a design execution.
    """

    uid = properties.UUID('id')
    """:UUID: unique external Citrine id of the material"""
    material_id = properties.UUID('material_id')
    """:UUID: unique internal Citrine id of the material"""
    identifiers = properties.List(properties.String(), 'identifiers')
    """:List[str]: globally unique identifiers assigned to the material"""
    primary_score = properties.Float('primary_score')
    """:float: numerical score describing how well the candidate satisfies the objectives
    and constraints (higher is better)"""
    material = properties.Object(DesignMaterial, 'material')
    """:DesignMaterial: the material returned by the design workflow"""


class HierarchicalDesignCandidate(Serializable["HierarchicalDesignCandidate"]):
    """A hierarchical candidate material generated by the Citrine Platform.

    This class represents the candidate computed by a design execution.
    """

    uid = properties.UUID('id')
    """:UUID: unique external Citrine ID of the material"""
    primary_score = properties.Float('primary_score')
    """:float: numerical score describing how well the candidate satisfies the objectives
    and constraints (higher is better)"""
    rank = properties.Integer("rank")
    """:int: rank by score amongst other candidates produced by the design workflow"""
    material = properties.Object(HierarchicalDesignMaterial, "material")
    """:HierarchicalDesignMaterial: the material returned by the design workflow"""


class SampleSearchSpaceResultCandidate(Serializable["SampleSearchSpaceResultCandidate"]):
    """A hierarchical candidate material generated by the Citrine Platform.

    This class represents the candidate computed by a design execution.
    """

    uid = properties.UUID('id')
    """:UUID: unique external Citrine ID of the material"""
    execution_uid = properties.UUID('id')
    """:UUID: unique external Citrine ID of the execution"""
    material = properties.Object(HierarchicalDesignMaterial, "material")
    """:HierarchicalDesignMaterial: the material returned by the design workflow"""



"""Resources that represent parameter templates."""
from typing import List, Dict, Optional, Type

from citrine._rest.resource import GEMDResource
from citrine.resources.attribute_templates import AttributeTemplate, AttributeTemplateCollection
from gemd.entity.bounds.base_bounds import BaseBounds
from gemd.entity.template.parameter_template import ParameterTemplate as GEMDParameterTemplate


class ParameterTemplate(
    GEMDResource['ParameterTemplate'],
    AttributeTemplate,
    GEMDParameterTemplate,
    typ=GEMDParameterTemplate.typ
):
    """
    A parameter template.

    Parameters
    ----------
    name: str
        The name of the parameter template.
    bounds: :py:class:`BaseBounds <gemd.entity.bounds.base_bounds.BaseBounds>`
        Bounds circumscribe the values that are valid according to this parameter template.
    description: str, optional
        A long-form description of the parameter template.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/gemd-docs/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/gemd-docs/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.

    """

    _response_key = GEMDParameterTemplate.typ  # 'parameter_template'

    def __init__(self,
                 name: str,
                 *,
                 bounds: BaseBounds,
                 uids: Optional[Dict[str, str]] = None,
                 description: Optional[str] = None,
                 tags: Optional[List[str]] = None):
        if uids is None:
            uids = dict()
        super(AttributeTemplate, self).__init__()
        GEMDParameterTemplate.__init__(self, name=name, bounds=bounds, tags=tags,
                                       uids=uids, description=description)

    def __str__(self):
        return '<Parameter template {!r}>'.format(self.name)


class ParameterTemplateCollection(AttributeTemplateCollection[ParameterTemplate]):
    """A collection of parameter templates."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/parameter-templates'
    _dataset_agnostic_path_template = 'projects/{project_id}/parameter-templates'
    _individual_key = 'parameter_template'
    _collection_key = 'parameter_templates'
    _resource = ParameterTemplate

    @classmethod
    def get_type(cls) -> Type[ParameterTemplate]:
        """Return the resource type in the collection."""
        return ParameterTemplate



import warnings
from typing import List, Optional

from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine.informatics.data_sources import DataSource
from citrine.informatics.descriptors import FormulationDescriptor, FormulationKey
from citrine.informatics.predictors import PredictorNode
from citrine.informatics.predictors.node import _check_deprecated_training_data

__all__ = ['SimpleMixturePredictor']


class SimpleMixturePredictor(Resource["SimpleMixturePredictor"], PredictorNode):
    """A predictor interface that flattens a formulation into a simple mixture.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        description of the predictor
    training_data: Optional[List[DataSource]]
        Sources of training data. Each can be either a CSV or an GEM Table. Candidates from
        multiple data sources will be combined into a flattened list and de-duplicated by uid and
        identifiers. De-duplication is performed if a uid or identifier is shared between two or
        more rows. The content of a de-duplicated row will contain the union of data across all
        rows that share the same uid or at least 1 identifier. Training data is unnecessary if the
        predictor is part of a graph that includes all training data required by this predictor.

    """

    training_data = properties.List(properties.Object(DataSource), 'training_data', default=[])

    typ = properties.String('type', default='SimpleMixture', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 input_descriptor: Optional[FormulationDescriptor] = None,
                 output_descriptor: Optional[FormulationDescriptor] = None,
                 training_data: Optional[List[DataSource]] = None):
        self.name: str = name
        self.description: str = description

        _check_deprecated_training_data(training_data)
        self.training_data: List[DataSource] = training_data or []

        if input_descriptor is not None:
            warnings.warn(
                "The field `input_descriptor` on a SimpleMixturePredictor is deprecated "
                "and will be ignored. The Citrine Platform will automatically generate a "
                f"FormulationDescriptor with key '{FormulationKey.HIERARCHICAL.value}' as input.",
                DeprecationWarning
            )

        if output_descriptor is not None:
            warnings.warn(
                "The field `output_descriptor` on a SimpleMixturePredictor is deprecated "
                "and will be ignored. The Citrine Platform will automatically generate a "
                f"FormulationDescriptor with key '{FormulationKey.FLAT.value}' as output.",
                DeprecationWarning
            )

    def __str__(self):
        return '<SimpleMixturePredictor {!r}>'.format(self.name)

    @property
    def input_descriptor(self) -> FormulationDescriptor:
        """The input formulation descriptor with key 'Formulation'."""
        return FormulationDescriptor.hierarchical()

    @property
    def output_descriptor(self) -> FormulationDescriptor:
        """The output formulation descriptor with key 'Flat Formulation'."""
        return FormulationDescriptor.flat()



from typing import Dict, List, Optional
from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from gemd.enumeration.base_enumeration import BaseEnumeration


class FingerprintType(BaseEnumeration):
    """[ALPHA] Fingerprint type used in Generative Design.

    * AP is the Atom Pairs Fingerprint.
    * PHCO is the Path-Length Connectivity Fingerprint.
    * BPF is the Binary Path Fingerprint.
    * PATH is the Paths of Atoms of Heteroatoms Fingerprint.
    * ECFP4 is the Extended Connectivity Fingerprint with radius 4.
    * ECFP6 is the Extended Connectivity Fingerprint with radius 6.
    * FCFP4 is the Focused Connectivity Fingerprint with radius 4.
    * FCFP6 is the Focused Connectivity Fingerprint with radius 6.
    """

    AP = "AP"
    PHCO = "PHCO"
    BPF = "BPF"
    PATH = "PATH"
    ECFP4 = "ECFP4"
    ECFP6 = "ECFP6"
    FCFP4 = "FCFP4"
    FCFP6 = "FCFP6"


class StructureExclusion(BaseEnumeration):
    """[ALPHA] Structure exclusion type used in Generative Design.

    * `DOUBLE_BONDS` excludes mutation steps that add double bonds.
    * `TRIPLE_BONDS` excludes mutation steps that add triple bonds.
    * `ANIONS` excludes mutation steps that add anions.
    * `CATIONS` excludes mutation steps that add cations.
    * `IONS` excludes mutation steps that add anions and cations.
    * `BORON` excludes mutation steps that add boron atoms.
    * `PHOSPHORUS` excludes mutation steps that add phosphorus atoms.
    * `SULFUR` excludes mutation steps that add sulfur atoms.
    * `NITROGEN` excludes mutation steps that add nitrogen atoms.
    * `OXYGEN` excludes mutation steps that add oxygen atoms.
    * `FLUORINE` excludes mutation steps that add fluorine atoms.
    * `BROMINE` excludes mutation steps that add bromine atoms.
    * `IODINE` excludes mutation steps that add iodine atoms.
    * `CHLORINE` excludes mutation steps that add chlorine atoms.
    """

    DOUBLE_BONDS = "DOUBLE_BONDS"
    TRIPLE_BONDS = "TRIPLE_BONDS"
    ANIONS = "ANIONS"
    CATIONS = "CATIONS"
    IONS = "IONS"
    BORON = "BORON"
    PHOSPHORUS = "PHOSPHORUS"
    SULFUR = "SULFUR"
    NITROGEN = "NITROGEN"
    OXYGEN = "OXYGEN"
    FLUORINE = "FLUORINE"
    BROMINE = "BROMINE"
    IODINE = "IODINE"
    CHLORINE = "CHLORINE"


class GenerativeDesignResult(Serializable["GenerativeDesignResult"]):
    """A Citrine Generation Design Execution Result.

    This class represents the result of a generative design execution.
    """

    @classmethod
    def _pre_build(cls, data: dict) -> dict:
        """Run data modification before building."""
        result = data.pop("result")
        data.update(result)
        return data

    uid = properties.UUID('id')
    execution_id = properties.UUID('execution_id')

    seed = properties.String("seed")
    """The seed used to generate the molecule."""
    mutated = properties.String("mutated")
    """The mutated molecule."""
    fingerprint_similarity = properties.Float("fingerprint_similarity")
    """The fingerprint similarity between the seed and the mutated molecule."""
    fingerprint_type = properties.String("fingerprint_type")
    """The fingerprint type used to calculate the fingerprint similarity."""

    def __init__(self):
        pass  # pragma: no cover


class GenerativeDesignInput(Serializable['GenerativeDesignInput']):
    """A Citrine Generative Design Execution Input.

    Parameters
    ----------
    seeds: List[str]
        The seeds used to generate the molecules.
    fingerprint_type: FingerprintType
        The fingerprint type used to calculate the fingerprint similarity.
    min_fingerprint_similarity: float
        The minimum fingerprint similarity between the seed and the mutated molecule.
    mutation_per_seed: int
        The number of initial mutations that will be attempted per seed.
        IMPORTANT, the total number of molecules generated will likely be lower than this value.
        Some mutations may be duplicates or may not meet the fingerprint similarity threshold.
    structure_exclusions: List[StructureExclusion]
        The structure exclusions used to limit molecule mutations.
        If None, no structure exclusions will be used.
    min_substructure_counts: Dict[str, int]
        Dictionary for constraining which substructures (represented by SMILES strings)
        must appear in each mutated molecule, along with integer values representing the
        minimum number of times each substructure must appear in a molecule to be
        considered a valid mutation.

    """

    seeds = properties.List(properties.String(), 'seeds')
    fingerprint_type = properties.Enumeration(FingerprintType, "fingerprint_type")
    min_fingerprint_similarity = properties.Float("min_fingerprint_similarity")
    mutation_per_seed = properties.Integer("mutation_per_seed")
    structure_exclusions = properties.List(
        properties.Enumeration(StructureExclusion),
        "structure_exclusions"
    )
    min_substructure_counts = properties.Mapping(
        properties.String(), properties.Integer(), "min_substructure_counts",
    )

    def __init__(
        self, *,
        seeds: List[str],
        fingerprint_type: FingerprintType,
        min_fingerprint_similarity: float,
        mutation_per_seed: int,
        structure_exclusions: Optional[List[StructureExclusion]] = None,
        min_substructure_counts: Optional[Dict[str, int]] = None,
    ):
        self.seeds: List[str] = seeds
        self.fingerprint_type: FingerprintType = fingerprint_type
        self.min_fingerprint_similarity: float = min_fingerprint_similarity
        self.mutation_per_seed: int = mutation_per_seed
        self.structure_exclusions: List[StructureExclusion] = structure_exclusions or []
        self.min_substructure_counts: Dict[str, int] = min_substructure_counts or {}



from typing import List, Optional
from uuid import UUID

from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine.informatics.design_candidate import DesignMaterial

__all__ = ['SinglePredictRequest']


class SinglePredictRequest(Serializable["SinglePredictRequest"]):
    """A Citrine Single Predict Request.

    This class represents a request to make a prediction against a predictor.
    """

    material_id = properties.UUID('material_id')
    identifiers = properties.List(properties.String(), 'identifiers')
    material = properties.Object(DesignMaterial, 'material')
    random_seed = properties.Optional(properties.Integer, 'random_seed')

    def __init__(self, material_id: UUID,
                 identifiers: List[str],
                 material: DesignMaterial,
                 *,
                 random_seed: Optional[int] = None):
        self.material_id = material_id
        self.identifiers = identifiers
        self.material = material
        self.random_seed = random_seed



from logging import getLogger
from math import isclose
from typing import Type, Union, List

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable

__all__ = ['PredictorEvaluationMetric',
           'RMSE',
           'NDME',
           'RSquared',
           'StandardRMSE',
           'PVA',
           'F1',
           'AreaUnderROC',
           'CoverageProbability']

logger = getLogger(__name__)


class PredictorEvaluationMetric(PolymorphicSerializable["PredictorEvaluationMetric"]):
    """A metric computed during a Predictor Evaluation Workflow.

    Abstract type that returns the proper type given a serialized dict.
    """

    def _attrs(self) -> List[str]:
        return ["typ"]

    def __hash__(self):
        return hash(self.__getattribute__("typ"))

    def __eq__(self, other):
        try:
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in self._attrs()
            ])
        except AttributeError:
            return False

    def __init__(self):
        pass

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        t = data["type"]
        return {
            "RMSE": RMSE,
            "RSquared": RSquared,
            "NDME": NDME,
            "StandardRMSE": StandardRMSE,
            "PVA": PVA,
            "F1": F1,
            "AreaUnderROC": AreaUnderROC,
            "CoverageProbability": CoverageProbability,
        }[t]


class RMSE(Serializable["RMSE"], PredictorEvaluationMetric):
    """Root-mean-square error."""

    typ = properties.String("type", default="RMSE", deserializable=False)

    def __repr__(self):
        return "rmse"

    def __str__(self):
        return "RMSE"


class NDME(Serializable["NDME"], PredictorEvaluationMetric):
    """Non-dimensional model error.

    The non-dimensional model error is the RMSE divided by the standard deviation
    of the labels in the training data (including all folds, not just the training folds).
    """

    typ = properties.String("type", default="NDME", deserializable=False)

    def __repr__(self):
        return "ndme"

    def __str__(self):
        return "NDME"


class RSquared(Serializable["RSquared"], PredictorEvaluationMetric):
    """Fraction of variance explained, commonly known as R^2.

    This dimensionless metric is equal to 1 - (mean squared error / variance of data).
    It is equal to the coefficient of determination calculated with respect to the line
    `predicted = actual`, hence it is commonly referred to as R^2. But unlike R^2 from
    ordinary linear regression, this metric can be negative.
    """

    typ = properties.String("type", default="RSquared", deserializable=False)

    def __repr__(self):
        return "R^2"

    def __str__(self):
        return "R^2"


class StandardRMSE(Serializable["StandardRMSE"], PredictorEvaluationMetric):
    """Standardized root-mean-square error."""

    typ = properties.String("type", default="StandardRMSE", deserializable=False)

    def __repr__(self):
        return "standardized_rmse"

    def __str__(self):
        return "Standardized RMSE"


class PVA(Serializable["PVA"], PredictorEvaluationMetric):
    """Predicted vs. actual data.

    Results are returned as a flattened list, where each item represents
    predicted vs. actual data for a single point.
    """

    typ = properties.String("type", default="PVA", deserializable=False)

    def __repr__(self):
        return "predicted_vs_actual"

    def __str__(self):
        return "Predicted vs Actual"


class F1(Serializable["F1"], PredictorEvaluationMetric):
    """Support-weighted F1 score."""

    typ = properties.String("type", default="F1", deserializable=False)

    def __repr__(self):
        return "f1"

    def __str__(self):
        return "F1 Score"


class AreaUnderROC(Serializable["AreaUnderROC"], PredictorEvaluationMetric):
    """Area under the receiver operating characteristic (ROC) curve."""

    typ = properties.String("type", default="AreaUnderROC", deserializable=False)

    def __repr__(self):
        return "area_under_roc"

    def __str__(self):
        return "Area Under the ROC"


class CoverageProbability(Serializable["CoverageProbability"], PredictorEvaluationMetric):
    """Percentage of observations that fall within a given confidence interval.

    The coverage level can be specified to 3 digits, e.g., 0.123, but not 0.1234.

    Parameters
    ----------
    coverage_level: Union[str, float]
        Confidence-interval coverage level.
        The coverage level must be between 0 and 1.0 (non-inclusive) and will be rounded
        to 3 significant figures.  Default: 0.683 corresponds to one std. deviation

    """

    _level_str = properties.String("coverage_level")
    typ = properties.String("type", default="CoverageProbability", deserializable=False)

    def __init__(self, *, coverage_level: Union[str, float] = "0.683"):
        if isinstance(coverage_level, str):
            try:
                raw_float = float(coverage_level)
            except ValueError:
                raise ValueError(
                    "Invalid coverage level string '{requested_level}'. "
                    "Coverage level must represent a floating point number between "
                    "0 and 1 (non-inclusive).".format(
                        requested_level=coverage_level
                    ))
        elif isinstance(coverage_level, float):
            raw_float = coverage_level
        else:
            raise TypeError("Coverage level must be a string or float")

        if raw_float >= 1.0 or raw_float <= 0.0:
            raise ValueError("Coverage level must be between 0 and 1 (non-inclusive).")
        _level_float = round(raw_float, 3)
        if not isclose(_level_float, raw_float):
            logger.warning(
                "Coverage level can only be specified to 3 decimal places."
                "Requested level '{requested_level}' will be rounded "
                "to {rounded_level}.".format(
                    requested_level=coverage_level,
                    rounded_level=_level_float
                ))

        self._level_str = "{:5.3f}".format(_level_float)

    def _attrs(self) -> List[str]:
        return ["typ", "_level_str"]

    def __repr__(self):
        return "coverage_probability_{}".format(self._level_str)

    def __str__(self):
        return "Coverage Probability ({})".format(self._level_str)



import warnings
from typing import Set, Mapping, Optional, Tuple, Union

from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine.informatics.constraints.constraint import Constraint
from citrine.informatics.descriptors import FormulationDescriptor

__all__ = ['IngredientRatioConstraint']


class IngredientRatioConstraint(Serializable['IngredientRatioConstraint'], Constraint):
    """A formulation constraint operating on the ratio of quantities of ingredients and a basis.

    Example: "6 to 7 parts ingredient A per 100 parts ingredient B" becomes
    .. code:: python
        IngredientRatioConstraint(min=6, max=7, ingredient=("A", 100), basis_ingredients=["B"])

    Parameters
    ----------
    formulation_descriptor: FormulationDescriptor
        descriptor to constrain
    min: float
        minimum value for the ratio
    max: float
        maximum value for the ratio
    ingredient: Optional[tuple[str, float]]
        multiplier for an ingredient in the numerator of the ratio
    label: Optional[tuple[str, float]]
        multiplier for a label in the numerator of the ratio
    basis_ingredients: Optional[Union[list[str], dict[str, float]]]
        the ingredients which should appear in the denominator of the ratio
    basis_labels: Optional[Union[list[str], dict[str, float]]]
        the labels which should appear in the denominator of the ratio

    """

    formulation_descriptor = properties.Object(FormulationDescriptor, 'formulation_descriptor')
    min = properties.Float('min')
    max = properties.Float('max')

    # The backend provides ingredients and labels as dictionaries, but presently only allows one
    # between them. To clarify customer interaction, we only allow a single one of each to be set.
    # Since our serde library doesn't allow extracting from a dict with unknown keys, we do it by
    # hiding the dictionaries and exposing properties.
    _ingredients = properties.Mapping(
        properties.String, properties.Float, 'ingredients', default={})
    _labels = properties.Mapping(properties.String, properties.Float, 'labels', default={})

    # The backend provides basis ingredients and basis labels as a dictionary from the key to a
    # multiplier. However, for ingredient ratio constraints, the multiplier in the denominator
    # should always be one, so we can't allow users to enter it. We need to use properties for this
    # behavior. It also allows us to display deprecation warnings for the coming type change.
    _basis_ingredients = properties.Mapping(
        properties.String, properties.Float, 'basis_ingredients', default={})
    _basis_labels = properties.Mapping(
        properties.String, properties.Float, 'basis_labels', default={})

    typ = properties.String('type', default='IngredientRatio')

    def __init__(self, *,
                 formulation_descriptor: FormulationDescriptor,
                 min: float,
                 max: float,
                 ingredient: Optional[Tuple[str, float]] = None,
                 label: Optional[Tuple[str, float]] = None,
                 basis_ingredients: Union[Set[str], Mapping[str, float]] = set(),
                 basis_labels: Union[Set[str], Mapping[str, float]] = set()):
        self.formulation_descriptor = formulation_descriptor
        self.min = min
        self.max = max
        self.ingredient = ingredient
        self.label = label
        self.basis_ingredients = basis_ingredients
        self.basis_labels = basis_labels

    @property
    def ingredient(self):
        """Retrieve the ingredient and its multiplier from the numerator, if it's been set."""
        return self._numerator_read(self._ingredients)

    @ingredient.setter
    def ingredient(self, value: Optional[Tuple[str, float]]):
        """Set the ingredient and its multiplier in the numerator."""
        self._ingredients = self._numerator_validate(value, "Ingredient")

    @property
    def label(self):
        """Retrieve the label and its multiplier from the numerator, if it's been set."""
        return self._numerator_read(self._labels)

    @label.setter
    def label(self, value: Optional[Tuple[str, float]]):
        """Set the label and its multiplier in the numerator."""
        self._labels = self._numerator_validate(value, "Label")

    @property
    def basis_ingredients(self) -> Mapping[str, float]:
        """Retrieve the ingredients in the denominator of the ratio."""
        warnings.warn("basis_ingredients is deprecated as of 2.13.0 and will change in 3.0. "
                      "Please use basis_ingredient_names instead.", DeprecationWarning)
        return self._basis_ingredients

    @basis_ingredients.setter
    def basis_ingredients(self, value: Union[Set[str], Mapping[str, float]]):
        """Set the ingredients in the denominator of the ratio."""
        if isinstance(value, dict):
            warnings.warn("As of version 2.13.0, multipliers for all basis ingredients are "
                          "ignored, so basis_ingredients should be a list of ingredient names.",
                          DeprecationWarning)
            value = set(value.keys())

        self.basis_ingredient_names = value

    @property
    def basis_ingredient_names(self) -> Set[str]:
        """Retrieve the names of all ingredients in the denominator of the ratio."""
        return set(self._basis_ingredients.keys())

    # This is for symmetry; it's not strictly necessary.
    @basis_ingredient_names.setter
    def basis_ingredient_names(self, value: Set[str]):
        """Set the names of all ingredients in the denominator of the ratio."""
        self._basis_ingredients = dict.fromkeys(value, 1)

    @property
    def basis_labels(self) -> Mapping[str, float]:
        """Retrieve the labels in the denominator of the ratio."""
        warnings.warn("basis_labels is deprecated as of 2.13.0 and will change in 3.0. Please use "
                      "basis_label_names instead.", DeprecationWarning)
        return self._basis_labels

    @basis_labels.setter
    def basis_labels(self, value: Union[Set[str], Mapping[str, float]]):
        """Set the labels in the denominator of the ratio."""
        if isinstance(value, dict):
            warnings.warn("As of version 2.13.0, multipliers for all basis labels are ignored, so "
                          "basis_labels should be a list of label names.", DeprecationWarning)
            value = set(value.keys())

        self.basis_label_names = value

    @property
    def basis_label_names(self) -> Set[str]:
        """Retrieve the names of all labels in the denominator of the ratio."""
        return set(self._basis_labels.keys())

    # This is for symmetry; it's not strictly necessary.
    @basis_label_names.setter
    def basis_label_names(self, value: Set[str]):
        """Set the names of all labels in the denominator of the ratio."""
        self._basis_labels = dict.fromkeys(value, 1)

    def _numerator_read(self, num_dict):
        if num_dict:
            return tuple(num_dict.items())[0]
        else:
            return None

    def _numerator_validate(self, value, input_type):
        if value:
            if len(value) != 2:
                raise ValueError(f"{input_type} must be a name and a multiplier.")

            name, mult = value
            if mult <= 0.0:
                raise ValueError(f"{input_type} multipler must be greater than 0.")

            return {name: mult}
        else:
            return {}

    def __str__(self):
        return f"<IngredientRatioConstraint '{self.formulation_descriptor.key}'>"



import csv
import json
from functools import partial
from io import StringIO
from typing import Dict, Iterator, Optional, Union
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine._session import Session
from citrine.informatics.experiment_values import ExperimentValue


class CandidateExperimentSnapshot(Serializable['CandidateExperimentSnapshot']):
    """The contents of a candidate experiment within an experiment data source."""

    uid = properties.UUID('experiment_id', serializable=False)
    """:UUID: unique Citrine id of this experiment"""
    candidate_id = properties.UUID('candidate_id', serializable=False)
    """:UUID: unique Citrine id of the candidate associated with this experiment"""
    workflow_id = properties.UUID('workflow_id', serializable=False)
    """:UUID: unique Citrine id of the design workflow which produced the associated candidate"""
    name = properties.String('name', serializable=False)
    """:str: name of the experiment"""
    description = properties.Optional(properties.String, 'description', serializable=False)
    """:Optional[str]: description of the experiment"""
    updated_time = properties.Datetime('updated_time', serializable=False)
    """:datetime: date and time at which the experiment was updated"""

    overrides = properties.Mapping(properties.String, properties.Object(ExperimentValue),
                                   'overrides')
    """:dict[str, ExperimentValue]: dictionary of candidate material variable overrides"""

    def __init__(self, *args, **kwargs):
        """Candidate experiment snapshots are not directly instantiated by the user."""
        pass  # pragma: no cover

    def _overrides_json(self) -> Dict[str, str]:
        return {name: json.dumps(expt_value.value) for name, expt_value in self.overrides.items()}


class ExperimentDataSource(Serializable['ExperimentDataSource']):
    """An experiment data source."""

    uid = properties.UUID('id', serializable=False)
    """:UUID: unique Citrine id of this experiment data source"""
    experiments = properties.List(properties.Object(CandidateExperimentSnapshot),
                                  'data.experiments',
                                  serializable=False)
    """:list[CandidateExperimentSnapshot]: list of experiment data in this data source"""
    branch_root_id = properties.UUID('metadata.branch_root_id', serializable=False)
    """:UUID: unique Citrine id of the branch root this data source is associated with"""
    version = properties.Integer('metadata.version', serializable=False)
    """:int: version of this data source"""
    created_by = properties.UUID('metadata.created.user', serializable=False)
    """:UUID: id of the user who created this data source"""
    create_time = properties.Datetime('metadata.created.time', serializable=False)
    """:datetime: date and time at which this data source was created"""

    def __init__(self, *args, **kwargs):
        """Experiment data sources are not directly instantiated by the user."""
        pass  # pragma: no cover

    def read(self) -> str:
        """Read this experiment data source into a CSV.

        Each row will be a single experiement from this data source, and each column is a variable
        which is overriden in any of the experiements in this data source. If an experiement did
        not override a variable, its cell will be left empty.

        All cells can be deserialized as JSON. Most of them will simply be strings or numbers. But
        if present,the "Formulation" cell will contain an escaped JSON string, which will
        deserialize into a mapping from ingredient names to floating-point values.
        """
        overrides = [experiment._overrides_json() for experiment in self.experiments]
        columns = {key for override in overrides for key in override.keys()}
        sorted_columns = sorted(list(columns), key=str.lower)

        buffer = StringIO()
        writer = csv.DictWriter(buffer, sorted_columns)
        writer.writeheader()
        writer.writerows(overrides)

        return buffer.getvalue()


class ExperimentDataSourceCollection(Collection[ExperimentDataSource]):
    """Represents the collection of all experiment data sources associated with a project."""

    _path_template = 'projects/{project_id}/candidate-experiment-datasources'
    _individual_key = None
    _resource = ExperimentDataSource
    _collection_key = 'response'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session

    def build(self, data: dict) -> ExperimentDataSource:
        """Build an individual experiment result from a dictionary."""
        result = ExperimentDataSource.build(data)
        result._project_id = self.project_id
        result._session = self.session
        return result

    def list(self, *,
             per_page: int = 100,
             branch_id: Optional[UUID] = None,
             version: Optional[Union[int, str]] = None) -> Iterator[ExperimentDataSource]:
        """Paginate over the experiment data sources.

        Parameters
        ---------
        per_page: int, optional
            Max number of results to return per page. Default is 100.  This parameter
            is used when making requests to the backend service.  If the page parameter
            is specified it limits the maximum number of elements in the response.
        branch_id: UUID, optional
            Filter the list by the root branch ID.
        version: Union[int, str], optional
            Filter the list by the data source version. Also accepts "latest".

        Returns
        -------
        Iterator[ExperimentDataSource]
            An iterator that can be used to loop over all matching experiment data sources.

        """
        params = {}
        if branch_id:
            params["branch"] = branch_id
        if version:
            params["version"] = version

        fetcher = partial(self._fetch_page, additional_params=params)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def read(self, datasource: Union[ExperimentDataSource, UUID, str]):
        """Reads the provided experiment data source into a CSV.

        If a UUID or str is provided, it's first retrieved from the platform.

        For details on the CSV format, see
        :py:meth:`citrine.resources.experiment_datasource.ExperimentDataSource.read`.
        """
        if not isinstance(datasource, ExperimentDataSource):
            datasource = self.get(uid=datasource)

        return datasource.read()



"""Tools for working with Scores."""
from typing import List, Optional

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine.informatics.constraints import Constraint
from citrine.informatics.objectives import Objective

__all__ = ['Score', 'LIScore', 'EIScore', 'EVScore']


class Score(PolymorphicSerializable['Score']):
    """A Score is used to rank materials according to objectives and constraints.

    Abstract type that returns the proper type given a serialized dict.

    """

    _name = properties.String('name')
    _description = properties.String('description')

    @classmethod
    def get_type(cls, data):
        """Return the subtype."""
        return {
            'MLI': LIScore,
            'MEI': EIScore,
            'MEV': EVScore
        }[data['type']]


class LIScore(Serializable['LIScore'], Score):
    """Evaluates the likelihood of scoring better than some baselines for given objectives.

    Parameters
    ----------
    objectives: list[Objective]
        objectives (e.g., maximize, minimize, tune, etc.)
        If multiple objectives are specified then this score evaluates the likelihood of
        simultaneously exceeding all objectives.
    baselines: list[float]
        best-so-far values for the various objectives (there must be one for each objective)
    constraints: list[Constraint]
        constraints limiting the allowed values that material instances can have

    """

    baselines = properties.List(properties.Float, 'baselines')
    objectives = properties.List(properties.Object(Objective), 'objectives')
    constraints = properties.List(properties.Object(Constraint), 'constraints')
    typ = properties.String('type', default='MLI')

    def __init__(self, *,
                 objectives: List[Objective],
                 baselines: List[float],
                 constraints: Optional[List[Constraint]] = None):
        self.objectives: List[Objective] = objectives
        self.baselines: List[float] = baselines
        self.constraints: List[Constraint] = constraints or []
        self._name = "Likelihood of Improvement"
        self._description = ""

    def __str__(self):
        return '<LIScore>'


class EIScore(Serializable['EIScore'], Score):
    """
    Evaluates the expected magnitude of improvement beyond baselines for a given objective.

    Parameters
    ----------
    objectives: list[Objective]
        objectives (e.g., maximize, minimize, tune, etc.)
        EIScore does not support more than 1 objective at this time.
    baselines: list[float]
        best-so-far values for the various objectives (there must be one for each objective)
    constraints: list[Constraint]
        constraints limiting the allowed values that material instances can have

    """

    baselines = properties.List(properties.Float, 'baselines')
    objectives = properties.List(properties.Object(Objective), 'objectives')
    constraints = properties.List(properties.Object(Constraint), 'constraints')
    typ = properties.String('type', default='MEI')

    def __init__(self, *,
                 objectives: List[Objective],
                 baselines: List[float],
                 constraints: Optional[List[Constraint]] = None):
        self.objectives: List[Objective] = objectives
        self.baselines: List[float] = baselines
        self.constraints: List[Constraint] = constraints or []
        self._name = "Expected Improvement"
        self._description = ""

    def __str__(self):
        return '<EIScore>'


class EVScore(Serializable['EVScore'], Score):
    """
    Evaluates the expected value for given objectives.

    Parameters
    ----------
    objectives: list[Objective]
        objectives (e.g., maximize, minimize, tune, etc.)
        If multiple objectives are specified, their scores are summed together. This allows
        for simultaneous optimization of multiple objectives, although the weighting of the
        various objectives cannot be directly specified.
    constraints: list[Constraint]
        constraints limiting the allowed values that material instances can have

    """

    objectives = properties.List(properties.Object(Objective), 'objectives')
    constraints = properties.List(properties.Object(Constraint), 'constraints')
    typ = properties.String('type', default='MEV')

    def __init__(self, *,
                 objectives: List[Objective],
                 constraints: Optional[List[Constraint]] = None):
        self.objectives: List[Objective] = objectives
        self.constraints: List[Constraint] = constraints or []
        self._name = "Expected Value"
        self._description = ""

    def __str__(self):
        return '<EVScore>'



"""Resources that represent collections of design spaces."""
import warnings
from typing import List, Optional, TypeVar, Union
from uuid import UUID

from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._utils.functions import format_escaped_url
from citrine.informatics.design_spaces import DesignSpace, EnumeratedDesignSpace, \
    HierarchicalDesignSpace
from citrine._rest.collection import Collection
from citrine._session import Session

CreationType = TypeVar('CreationType', bound=DesignSpace)


class DefaultDesignSpaceMode(BaseEnumeration):
    """The type of default design space to create.

    * ATTRIBUTE results in a product design space containing dimensions required by the predictor
    * HIERARCHICAL results in a hierarchical design space resembling the shape of training data
    """

    ATTRIBUTE = 'ATTRIBUTE'
    HIERARCHICAL = 'HIERARCHICAL'


class DesignSpaceCollection(Collection[DesignSpace]):
    """Represents the collection of design spaces as well as the resources belonging to it.

    Parameters
    ----------
    project_id: UUID
        the UUID of the project

    """

    _api_version = 'v3'
    _path_template = '/projects/{project_id}/design-spaces'
    _individual_key = None
    _resource = DesignSpace
    _collection_key = 'response'
    _enumerated_cell_limit = 128 * 2000

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session

    def build(self, data: dict) -> DesignSpace:
        """Build an individual design space."""
        design_space: DesignSpace = DesignSpace.build(data)
        design_space._session = self.session
        design_space._project_id = self.project_id
        return design_space

    def _verify_write_request(self, design_space: DesignSpace):
        """Perform write-time validations of the design space registration or update.

        EnumeratedDesignSpaces can be pretty big, so we want to return a helpful error message
        rather than let the POST or PUT call fail because the request body is too big.  This
        validation is performed when the design space is sent to the platform in case a user
        creates a large intermediate design space but then filters it down before registering it.
        """
        if isinstance(design_space, EnumeratedDesignSpace):
            width = len(design_space.descriptors)
            length = len(design_space.data)
            if width * length > self._enumerated_cell_limit:
                msg = "EnumeratedDesignSpace only supports up to {} descriptor-values, " \
                      "but {} were given. Please reduce the number of descriptors or candidates " \
                      "in this EnumeratedDesignSpace"
                raise ValueError(msg.format(self._enumerated_cell_limit, width * length))

    def _hydrate_design_space(self, design_space: DesignSpace) -> List[DesignSpace]:
        if design_space.typ != "ProductDesignSpace":
            return design_space

        subspaces = []
        for subspace in design_space.subspaces:
            if isinstance(subspace, (str, UUID)):
                warnings.warn("Support for UUIDs in subspaces is deprecated as of 2.16.0, and "
                              "will be dropped in 3.0. Please use DesignSpace objects instead.",
                              DeprecationWarning)
                subspaces.append(self.get(subspace))
            else:
                subspaces.append(subspace)
        design_space.subspaces = subspaces
        return design_space

    def register(self, design_space: DesignSpace) -> DesignSpace:
        """Create a new design space."""
        self._verify_write_request(design_space)
        hydrated_ds = self._hydrate_design_space(design_space)

        registered_ds = super().register(hydrated_ds)

        # If the initial response is invalid, just return it.
        # If not, kick off validation, since we never exposed saving a design space without
        # validation so we should continue to do it automatically
        if registered_ds.failed():
            return registered_ds
        else:
            return self._validate(registered_ds.uid)

    def update(self, design_space: DesignSpace) -> DesignSpace:
        """Update and validate an existing DesignSpace."""
        self._verify_write_request(design_space)
        hydrated_ds = self._hydrate_design_space(design_space)
        updated_ds = super().update(hydrated_ds)

        # The /api/v3/design-spaces endpoint switched archiving from a field on the update payload
        # to their own endpoints. To maintain backwards compatibility, all design spaces have an
        # _archived field set by the archived property. It will be archived if True, and restored
        # if False. It defaults to None, which does nothing. The value is reset afterwards.
        if design_space._archived is True:
            self.archive(design_space.uid)
        elif design_space._archived is False:
            self.restore(design_space.uid)
        design_space._archived = None

        # If the initial response is invalid, just return it.
        # If not, kick off validation, since we never exposed saving a design space without
        # validation so we should continue to do it automatically
        if updated_ds.failed():
            return updated_ds
        else:
            return self._validate(updated_ds.uid)

    def _validate(self, uid: Union[UUID, str]) -> DesignSpace:
        path = self._get_path(uid, action="validate")
        entity = self.session.put_resource(path, {}, version=self._api_version)
        return self.build(entity)

    def archive(self, uid: Union[UUID, str]) -> DesignSpace:
        """Archiving a design space removes it from view, but is not a hard delete.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the design space to archive

        """
        url = self._get_path(uid, action="archive")
        entity = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(entity)

    def restore(self, uid: Union[UUID, str]) -> DesignSpace:
        """Restore an archived design space.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the design space to restore

        """
        url = self._get_path(uid, action="restore")
        entity = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(entity)

    def create_default(self,
                       *,
                       predictor_id: UUID,
                       predictor_version: Optional[Union[int, str]] = None,
                       mode: DefaultDesignSpaceMode = DefaultDesignSpaceMode.ATTRIBUTE,
                       include_ingredient_fraction_constraints: bool = False,
                       include_label_fraction_constraints: bool = False,
                       include_label_count_constraints: bool = False,
                       include_parameter_constraints: bool = False) -> DesignSpace:
        """Create a default design space for a predictor.

        This method will return an unregistered design space for all inputs
        that are not responses of the predictor. The design space will contain a
        :class:`~citrine.informatics.design_spaces.formulation_design_space.FormulationDesignSpace`
        for each formulation input. Dimensions are constructed for all other inputs.
        A :class:`~citrine.informatics.dimensions.ContinuousDimension` is constructed for each
        input that corresponds to a :class:`~citrine.informatics.descriptors.RealDescriptor`.
        An :class:`~citrine.informatics.dimensions.EnumeratedDimension` is constructed for all
        other inputs. Enumerated values taken from the allowed ``categories`` of a
        :class:`~citrine.informatics.descriptors.CategoricalDescriptor` or from the
        unique values from the training data for all other descriptors.

        Parameters
        ----------
        predictor_id: UUID
            UUID of the predictor used to construct the design space

        predictor_version: Optional[Union[int, str]]
            Version of the predictor used to construct the design space

        mode: DefaultDesignSpaceMode
            The type of default design space to produce.
            Defaults to DefaultDesignSpaceMode.ATTRIBUTE.

        include_ingredient_fraction_constraints: bool
            Whether to include constraints on ingredient fractions based on the training data.
            Defaults to False.

        include_label_fraction_constraints: bool
            Whether to include constraints on label fractions based on the training data.
            Defaults to False.

        include_label_count_constraints: bool
            Whether to include constraints on labeled ingredient counts based on the training data.
            Defaults to False.

        include_parameter_constraints: bool
            Whether to include constraints on all other inputs based on the training data.
            Defaults to False.

        Returns
        -------
        DesignSpace
            Default design space

        """
        path = f'projects/{self.project_id}/design-spaces/default'
        payload = {
            "predictor_id": str(predictor_id),
            "mode": mode.value,
            "include_ingredient_fraction_constraints": include_ingredient_fraction_constraints,
            "include_label_fraction_constraints": include_label_fraction_constraints,
            "include_label_count_constraints": include_label_count_constraints,
            "include_parameter_constraints": include_parameter_constraints
        }
        if predictor_version:
            payload["predictor_version"] = predictor_version

        data = self.session.post_resource(path, json=payload, version=self._api_version)
        return self.build(DesignSpace.wrap_instance(data["instance"]))

    def convert_to_hierarchical(
            self,
            uid: Union[UUID, str],
            *,
            predictor_id: Union[UUID, str],
            predictor_version: Optional[Union[int, str]] = None
    ) -> HierarchicalDesignSpace:
        """Convert an existing ProductDesignSpace into an equivalent HierarchicalDesignSpace.

        A :class:`~citrine.informatics.design_spaces.ProductDesignSpace` can be mapped to a
        :class:`~citrine.informatics.design_spaces.HierarchicalDesignSpace` by using the associated
        predictor and its training data to infer the shape of the hierarchical design space.
        Constraints are then transferred from the product design space to the appropriate node
        in the hierarchical design space.

        Parameters
        ----------
        uid: Union[str, UUID]
            UUID of the existing product design space to convert to a hierarchical version
        predictor_id: Union[UUID, str]
            UUID of a predictor associated with the design space.
        predictor_version: Optional[Union[int, str]]
            Version of the predictor to use. Defaults to the most recent version.

        Returns
        -------
        A :class:`~citrine.informatics.design_spaces.HierarchicalDesignSpace` modeled on the
        existing :class:`~citrine.informatics.design_spaces.ProductDesignSpace`.

        """
        path = format_escaped_url(
            "projects/{project_id}/design-spaces/{design_space_id}/convert-hierarchical",
            project_id=self.project_id,
            design_space_id=uid
        )
        payload = {
            "predictor_id": str(predictor_id),
        }
        if predictor_version:
            payload["predictor_version"] = predictor_version
        data = self.session.post_resource(path, json=payload, version=self._api_version)
        return HierarchicalDesignSpace.build(DesignSpace.wrap_instance(data["instance"]))

    def delete(self, uid: Union[UUID, str]):
        """Design Spaces cannot be deleted at this time."""
        msg = "Design Spaces cannot be deleted at this time. Use 'archive' instead."
        raise NotImplementedError(msg)



"""Resources that represent ingredient spec data objects."""
from typing import List, Dict, Optional, Type, Iterator, Union
from uuid import UUID

from citrine._rest.resource import GEMDResource
from citrine._serialization.properties import List as PropertyList
from citrine._serialization.properties import String, LinkOrElse, Object
from citrine._serialization.properties import Optional as PropertyOptional
from citrine.resources.object_specs import ObjectSpec, ObjectSpecCollection
from gemd.entity.file_link import FileLink
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.object.ingredient_spec import IngredientSpec as GEMDIngredientSpec
from gemd.entity.object.material_spec import MaterialSpec as GEMDMaterialSpec
from gemd.entity.object.process_spec import ProcessSpec as GEMDProcessSpec
from gemd.entity.value.continuous_value import ContinuousValue


class IngredientSpec(
    GEMDResource['IngredientSpec'],
    ObjectSpec,
    GEMDIngredientSpec,
    typ=GEMDIngredientSpec.typ
):
    """
    An ingredient specification.

    Ingredients annotate a material with information about its usage in a process.

    Parameters
    ----------
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/gemd-docs/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/gemd-docs/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the ingredient spec.
    material: MaterialSpec
        Material that this ingredient is.
    process: ProcessSpec
        Process that this ingredient is used in.
    mass_fraction: :py:class:`ContinuousValue \
    <gemd.entity.value.continuous_value.ContinuousValue>`, optional
        The mass fraction of the ingredient in the process.
    volume_fraction: :py:class:`ContinuousValue \
    <gemd.entity.value.continuous_value.ContinuousValue>`, optional
        The volume fraction of the ingredient in the process.
    number_fraction: :py:class:`ContinuousValue \
    <gemd.entity.value.continuous_value.ContinuousValue>`, optional
        The number fraction of the ingredient in the process.
    absolute_quantity: :py:class:`ContinuousValue \
    <gemd.entity.value.continuous_value.ContinuousValue>`, optional
        The absolute quantity of the ingredient in the process.
    name: str
        Label on the ingredient that is unique within the process that contains it.
    labels: List[str], optional
        Additional labels on the ingredient.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    """

    _response_key = GEMDIngredientSpec.typ  # 'ingredient_spec'

    material = PropertyOptional(LinkOrElse(GEMDMaterialSpec), 'material', override=True)
    process = PropertyOptional(LinkOrElse(GEMDProcessSpec),
                               'process',
                               override=True,
                               use_init=True)
    mass_fraction = PropertyOptional(Object(ContinuousValue), 'mass_fraction', override=True)
    volume_fraction = PropertyOptional(Object(ContinuousValue), 'volume_fraction', override=True)
    number_fraction = PropertyOptional(Object(ContinuousValue), 'number_fraction', override=True)
    absolute_quantity = PropertyOptional(Object(ContinuousValue),
                                         'absolute_quantity',
                                         override=True)
    name = String('name', override=True, use_init=True)
    labels = PropertyOptional(PropertyList(String()), 'labels', override=True, use_init=True)

    def __init__(self,
                 name: str,
                 *,
                 uids: Optional[Dict[str, str]] = None,
                 tags: Optional[List[str]] = None,
                 notes: Optional[str] = None,
                 material: Optional[GEMDMaterialSpec] = None,
                 process: Optional[GEMDProcessSpec] = None,
                 mass_fraction: Optional[ContinuousValue] = None,
                 volume_fraction: Optional[ContinuousValue] = None,
                 number_fraction: Optional[ContinuousValue] = None,
                 absolute_quantity: Optional[ContinuousValue] = None,
                 labels: Optional[List[str]] = None,
                 file_links: Optional[List[FileLink]] = None):
        if uids is None:
            uids = dict()

        super(ObjectSpec, self).__init__()
        GEMDIngredientSpec.__init__(self, uids=uids, tags=tags, notes=notes,
                                    material=material, process=process,
                                    mass_fraction=mass_fraction, volume_fraction=volume_fraction,
                                    number_fraction=number_fraction,
                                    absolute_quantity=absolute_quantity, labels=labels,
                                    name=name, file_links=file_links)

    def __str__(self):
        return '<Ingredient spec {!r}>'.format(self.name)


class IngredientSpecCollection(ObjectSpecCollection[IngredientSpec]):
    """Represents the collection of all ingredient specs associated with a dataset."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/ingredient-specs'
    _dataset_agnostic_path_template = 'projects/{project_id}/ingredient-specs'
    _individual_key = 'ingredient_spec'
    _collection_key = 'ingredient_specs'
    _resource = IngredientSpec

    @classmethod
    def get_type(cls) -> Type[IngredientSpec]:
        """Return the resource type in the collection."""
        return IngredientSpec

    def list_by_process(self,
                        uid: Union[UUID, str, LinkByUID, GEMDProcessSpec]
                        ) -> Iterator[IngredientSpec]:
        """
        Get ingredients to a process.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDProcessSpec]
            A representation of the process spec whose ingredients are to be located.

        Returns
        -------
        Iterator[IngredientSpec]
            The ingredients to the specified process.

        """
        return self._get_relation(relation='process-specs', uid=uid)

    def list_by_material(self,
                         uid: Union[UUID, str, LinkByUID, GEMDMaterialSpec]
                         ) -> Iterator[IngredientSpec]:
        """
        Get ingredients using the specified material.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDMaterialSpec]
            A representation of the material spec whose ingredient usages are to be located.

        Returns
        -------
        Iterator[IngredientSpec]
            The ingredients using the specified material

        """
        return self._get_relation(relation='material-specs', uid=uid)



from abc import ABC
from typing import List, Optional
from uuid import UUID

from citrine._rest.asynchronous_object import AsynchronousObject
from citrine._rest.pageable import Pageable
from citrine._rest.paginator import Paginator
from citrine._serialization import properties
from citrine._session import Session
from citrine.resources.status_detail import StatusDetail


from deprecation import deprecated


class Execution(Pageable, AsynchronousObject, ABC):
    """A base class for execution resources.

    This class provides an abstraction for the execution resources.
    """

    _paginator: Paginator = Paginator()
    _collection_key = 'response'
    _in_progress_statuses = ["INPROGRESS"]
    _succeeded_statuses = ["SUCCEEDED"]
    _failed_statuses = ["FAILED"]
    _session: Optional[Session] = None
    project_id: Optional[UUID] = None

    uid: UUID = properties.UUID('id', serializable=False)
    """:UUID: Unique identifier of the execution"""
    status = properties.Optional(properties.String(), 'status', serializable=False)
    """:Optional[str]: short description of the execution's status"""
    status_description = properties.Optional(
        properties.String(), 'status_description', serializable=False)
    """:Optional[str]: more detailed description of the execution's status"""
    status_detail = properties.List(
        properties.Object(StatusDetail), 'status_detail', default=[], serializable=False
    )
    """:List[StatusDetail]: a list of structured status info, containing the message and level"""
    created_by = properties.Optional(properties.UUID, 'created_by', serializable=False)
    """:Optional[UUID]: id of the user who created the resource"""
    updated_by = properties.Optional(properties.UUID, 'updated_by', serializable=False)
    """:Optional[UUID]: id of the user who most recently updated the resource,
    if it has been updated"""
    create_time = properties.Optional(properties.Datetime, 'create_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was created"""
    update_time = properties.Optional(properties.Datetime, 'update_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was most recently updated,
    if it has been updated"""

    @property
    @deprecated(deprecated_in="2.2.0", removed_in="3.0.0", details="Use status_detail instead.")
    def status_info(self) -> List[str]:
        """:List[str]: human-readable explanations of the status."""
        return [detail.msg for detail in self.status_detail]

    def __str__(self):
        return f'<{self.__class__.__name__} {str(self.uid)!r}>'

    def _path(self):
        raise NotImplementedError("Subclasses must implement the _path method")  # pragma: no cover



import warnings
from typing import Set, Mapping, Optional

from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine.informatics.descriptors import FormulationDescriptor, RealDescriptor, FormulationKey
from citrine.informatics.predictors import PredictorNode

__all__ = ['IngredientsToFormulationPredictor']


class IngredientsToFormulationPredictor(
    Resource["IngredientsToFormulationPredictor"], PredictorNode
):
    """A predictor interface that constructs a formulation from ingredient quantities.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        description of the predictor
    id_to_quantity: Mapping[str, RealDescriptor]
        Map from ingredient identifier to the descriptor that represents its quantity,
        e.g., ``{'water': RealDescriptor('water quantity', 0, 1, "")}``
    labels: Mapping[str, Set[str]]
        Map from each label to all ingredients assigned that label, when present in a mixture,
        e.g., ``{'solvent': {'water'}}``

    """

    id_to_quantity = properties.Mapping(
        properties.String, properties.Object(RealDescriptor), 'id_to_quantity'
    )
    labels = properties.Mapping(properties.String, properties.Set(properties.String), 'labels')

    typ = properties.String('type', default='IngredientsToSimpleMixture', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 output: Optional[FormulationDescriptor] = None,
                 id_to_quantity: Mapping[str, RealDescriptor],
                 labels: Mapping[str, Set[str]]):
        self.name: str = name
        self.description: str = description
        self.id_to_quantity: Mapping[str, RealDescriptor] = id_to_quantity
        self.labels: Mapping[str, Set[str]] = labels

        if output is not None:
            warnings.warn(
                "The field `output` on an IngredientsToFormulationPredictor is deprecated "
                "and will be ignored. The Citrine Platform will automatically generate a "
                f"FormulationDescriptor with key '{FormulationKey.HIERARCHICAL.value}' as output.",
                DeprecationWarning
            )

    def __str__(self):
        return '<IngredientsToFormulationPredictor {!r}>'.format(self.name)

    @property
    def output(self) -> FormulationDescriptor:
        """The output formulation descriptor with key 'Formulation'."""
        return FormulationDescriptor.hierarchical()



from typing import Optional, Iterator, Iterable
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine.exceptions import CitrineException, BadRequest
from citrine.jobs.job import JobSubmissionResponse, JobFailureError, _poll_for_job_completion
from citrine.resources.api_error import ApiError, ValidationError
from gemd.enumeration.base_enumeration import BaseEnumeration


class IngestionStatusType(BaseEnumeration):
    """[ALPHA] State of the ingestion process."""

    INGESTION_CREATED = "ingestion_created"


class IngestionErrorFamily(BaseEnumeration):
    """[ALPHA] What class of ingest error was encountered."""

    FILE = "file"
    STRUCTURE = "structure"
    DATA = "data"
    UNKNOWN = "unknown"


class IngestionErrorType(BaseEnumeration):
    """[ALPHA] What ingest error was encountered."""

    FILE_EXTENSION_NOT_SUPPORTED = "file_extension_not_supported"
    MISSING_TYPE_HEADER = "missing_type_header"
    MISSING_RAW_FOR_INGREDIENT = "missing_raw_for_ingredient"
    DUPLICATED_MATERIAL = "duplicated_material"
    REGISTERING_OBJECTS_ERROR = "registering_objects_error"
    MISSING_ASPECT_TYPE = "missing_aspect_type"
    INVALID_DUPLICATE_NAME = "invalid_duplicate_name"
    INVALID_UNITS_ON_ASPECT = "invalid_units_on_aspect"
    INVALID_BASIS_ON_ASPECT = "invalid_basis_on_aspect"
    INVALID_FRACTION_ON_ASPECT = "invalid_fraction_on_aspect"
    INVALID_TYPE_HINT_ON_ASPECT = "invalid_type_hint_on_aspect"
    CATEGORICAL_OUTSIDE_BOUNDS_ERROR = "categorical_outside_bounds_error"
    INTEGER_OUTSIDE_BOUNDS_ERROR = "integer_outside_bounds_error"
    REAL_OUTSIDE_BOUNDS_ERROR = "real_outside_bounds_error"
    INVALID_PROCESS_REFERENCE = "invalid_process_reference"
    COLLIDING_ID = "colliding_id"
    UNKNOWN_ERROR = "unknown_error"


class IngestionErrorLevel(BaseEnumeration):
    """[ALPHA] Severity of the issue encountered."""

    ERROR = "error"
    WARNING = "warning"
    INFO = "info"


class IngestionErrorTrace(Resource['IngestionErrorTrace']):
    """[ALPHA] Detailed information about an ingestion issue."""

    family = properties.Enumeration(IngestionErrorFamily, "family")
    error_type = properties.Enumeration(IngestionErrorType, "error_type")
    level = properties.Enumeration(IngestionErrorLevel, "level")
    msg = properties.String("msg")
    dataset_file_id = properties.Optional(properties.UUID(), "dataset_file_id", default=None)
    file_version_uuid = properties.Optional(properties.UUID(), "file_version_uuid", default=None)
    row_number = properties.Optional(properties.Integer(), "row_number", default=None)
    column_number = properties.Optional(properties.Integer(), "column_number", default=None)

    def __init__(self,
                 msg,
                 level=IngestionErrorLevel.ERROR,
                 *,
                 family=IngestionErrorFamily.UNKNOWN,
                 error_type=IngestionErrorType.UNKNOWN_ERROR,
                 dataset_file_id=dataset_file_id.default,
                 file_version_uuid=file_version_uuid.default,
                 row_number=row_number.default,
                 column_number=column_number.default,
                 ):
        self.msg = msg
        self.level = level
        self.family = family
        self.error_type = error_type
        self.dataset_file_id = dataset_file_id
        self.file_version_uuid = file_version_uuid
        self.row_number = row_number
        self.column_number = column_number

    @classmethod
    def from_validation_error(cls, source: ValidationError) -> "IngestionErrorTrace":
        """[ALPHA] Generate an IngestionErrorTrace from a ValidationError."""
        return cls(
            msg=source.failure_message,
            level=IngestionErrorLevel.ERROR,
        )

    def __str__(self):
        return f"{self!r}: {self.msg}"

    def __repr__(self):
        coords = ", ".join([x for x in (self.column_number, self.row_number) if x is not None])
        return f"<{self.level}: {self.error_type}{f' {coords}' if len(coords) else ''}>"


class IngestionException(CitrineException):
    """
    [ALPHA] An exception that contains details of a failed ingestion.

    Attributes
    ----------
    uid: Optional[UUID]
    errors: List[IngestionErrorTrace]

    """

    uid = properties.Optional(properties.UUID(), 'ingestion_id', default=None)
    status = properties.Enumeration(IngestionStatusType, "status")
    errors = properties.List(properties.Object(IngestionErrorTrace), "errors")

    def __init__(self,
                 *,
                 uid: Optional[UUID] = uid.default,
                 errors: Iterable[IngestionErrorTrace]):
        errors_ = list(errors)
        message = '; '.join(str(e) for e in errors_)
        super().__init__(message)
        self.uid = uid
        self.errors = errors_

    @classmethod
    def from_status(cls, source: "IngestionStatus") -> "IngestionException":
        """[ALPHA] Build an IngestionException from an IngestionStatus."""
        return cls(uid=source.uid, errors=source.errors)

    @classmethod
    def from_api_error(cls, source: ApiError) -> "IngestionException":
        """[ALPHA] Build an IngestionException from an ApiError."""
        if len(source.validation_errors) > 0:
            return cls(errors=[IngestionErrorTrace.from_validation_error(x)
                               for x in source.validation_errors])
        else:
            return cls(errors=[IngestionErrorTrace(msg=source.message)])


class IngestionStatus(Resource['IngestionStatus']):
    """
    [ALPHA] An object that represents the outcome of an ingestion event.

    Attributes
    ----------
    uid: String
    status: IngestionStatusType
    errors: List[IngestionErrorTrace]

    """

    uid = properties.Optional(properties.UUID(), 'ingestion_id', default=None)
    status = properties.Enumeration(IngestionStatusType, "status")
    errors = properties.List(properties.Object(IngestionErrorTrace), "errors")

    def __init__(self,
                 *,
                 uid: Optional[UUID] = uid.default,
                 status: IngestionStatusType = IngestionStatusType.INGESTION_CREATED,
                 errors: Iterable[IngestionErrorTrace]):
        self.uid = uid
        self.status = status
        self.errors = list(errors)

    @property
    def success(self) -> bool:
        """Whether the Ingestion operation was error-free."""
        return len(self.errors) == 0

    @classmethod
    def from_exception(cls, exception: IngestionException) -> "IngestionStatus":
        """[ALPHA] Build an IngestionStatus from an IngestionException."""
        return cls(uid=exception.uid, errors=exception.errors)


class Ingestion(Resource['Ingestion']):
    """
    [ALPHA] A job that uploads new information to the platform.

    Datasets are the basic unit of access control. A user with read access to a dataset can view
    every object in that dataset. A user with write access to a dataset can create, update,
    and delete objects in the dataset.

    Attributes
    ----------
    uid: UUID
        Unique uuid4 identifier of this ingestion.

    """

    uid = properties.UUID('ingestion_id')
    project_id = properties.UUID('project_id')
    dataset_id = properties.UUID('dataset_id')
    session = properties.Object(Session, 'session', serializable=False)
    raise_errors = properties.Optional(properties.Boolean(), 'raise_errors', default=True)

    def build_objects(self,
                      *,
                      build_table: bool = False,
                      delete_dataset_contents: bool = False,
                      delete_templates: bool = True,
                      timeout: float = None,
                      polling_delay: Optional[float] = None
                      ) -> IngestionStatus:
        """
        [ALPHA] Perform a complete ingestion operation, from start to finish.

        Initiates an ingestion operation, polls the server to determine when the job
        has finished, and returns the outcome.

        Parameters
        ----------
        build_table: bool
            Whether to build a table immediately after ingestion.  Default : False
        delete_dataset_contents: bool
            Whether to delete objects prior to generating new gemd objects.  Default: False.
        delete_templates: bool
            Whether to delete all objects and templates (as opposed to not deleting
            templates) when `delete_dataset_contents` is True.  Default: True
        timeout: Optional[float]
            Amount of time to wait on the job (in seconds) before giving up. Note that
            this number has no effect on the underlying job itself, which can also time
            out server-side.
        polling_delay: Optional[float]
            How long to delay between each polling retry attempt.

        Returns
        ----------
        JobSubmissionResponse
            The object for the submitted job

        """
        try:
            job = self.build_objects_async(build_table=build_table,
                                           delete_dataset_contents=delete_dataset_contents,
                                           delete_templates=delete_templates)
        except IngestionException as e:
            if self.raise_errors:
                raise e
            else:
                return IngestionStatus.from_exception(e)

        status = self.poll_for_job_completion(job, timeout=timeout, polling_delay=polling_delay)

        if self.raise_errors and not status.success:
            raise IngestionException.from_status(status)

        return status

    def build_objects_async(self,
                            *,
                            build_table: bool = False,
                            delete_dataset_contents: bool = False,
                            delete_templates: bool = True) -> JobSubmissionResponse:
        """
        [ALPHA] Begin an async ingestion operation.

        Parameters
        ----------
        build_table: bool
            Whether to build a table immediately after ingestion.  Default : False
        delete_dataset_contents: bool
            Whether to delete objects prior to generating new gemd objects.  Default: False.
        delete_templates: bool
            Whether to delete all objects and templates (as opposed to not deleting
            templates) when `delete_dataset_contents` is True.  Default: True

        Returns
        ----------
        JobSubmissionResponse
            The object for the submitted job

        """
        collection = IngestionCollection(project_id=self.project_id,
                                         dataset_id=self.dataset_id,
                                         session=self.session)
        path = collection._get_path(uid=self.uid, action="gemd-objects-async")
        params = {
            "build_table": build_table,
            "delete_dataset_contents": delete_dataset_contents,
            "delete_templates": delete_templates,
        }
        try:
            return JobSubmissionResponse.build(
                self.session.post_resource(path=path, json={}, params=params)
            )
        except BadRequest as e:
            if e.api_error is not None:
                raise IngestionException.from_api_error(e.api_error)
            else:
                raise e

    def poll_for_job_completion(self,
                                job: JobSubmissionResponse,
                                *,
                                timeout: Optional[float] = None,
                                polling_delay: Optional[float] = None
                                ) -> IngestionStatus:
        """
        [ALPHA] Repeatedly ask server if a job associated with this ingestion has completed.

        Parameters
        ----------
        job: JobSubmissionResponse
            The job to check on
        timeout
            Amount of time to wait on the job (in seconds) before giving up. Note that
            this number has no effect on the underlying job itself, which can also time
            out server-side.
        polling_delay:
            How long to delay between each polling retry attempt.

        Returns
        ----------
        IngestionStatus
            A string representation of the status


        """
        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout
        if polling_delay is not None:
            kwargs["polling_delay"] = polling_delay

        _poll_for_job_completion(
            session=self.session,
            project_id=self.project_id,
            job=job,
            raise_errors=False,  # JobFailureError doesn't contain the error
            **kwargs
        )
        return self.status()

    def status(self) -> IngestionStatus:
        """
        [ALPHA] Retrieve the status of the ingestion  from platform.

        Returns
        ----------
        IngestionStatus
            The result of the ingestion attempt

        """
        collection = IngestionCollection(project_id=self.project_id,
                                         dataset_id=self.dataset_id,
                                         session=self.session)
        path = collection._get_path(uid=self.uid, action="status")
        return IngestionStatus.build(self.session.get_resource(path=path))


class FailedIngestion(Ingestion):
    """[ALPHA] Object to fill in when building an ingest fails."""

    def __init__(self, errors: Iterable[IngestionErrorTrace]):
        self.errors = list(errors)
        self.raise_errors = False

    def build_objects(self,
                      *,
                      build_table: bool = False,
                      delete_dataset_contents: bool = False,
                      delete_templates: bool = True) -> IngestionStatus:
        """[ALPHA] Satisfy the required interface for a failed ingestion."""
        return self.status()

    def build_objects_async(self,
                            *,
                            build_table: bool = False,
                            delete_dataset_contents: bool = False,
                            delete_templates: bool = True) -> JobSubmissionResponse:
        """[ALPHA] Satisfy the required interface for a failed ingestion."""
        raise JobFailureError(
            message=f"Errors: {[e.msg for e in self.errors]}",
            job_id=None,
            failure_reasons=[e.msg for e in self.errors]
        )

    def poll_for_job_completion(self,
                                job: JobSubmissionResponse,
                                *,
                                timeout: Optional[float] = None,
                                polling_delay: Optional[float] = None
                                ) -> IngestionStatus:
        """[ALPHA] Satisfy the required interface for a failed ingestion."""
        raise JobFailureError(
            message=f"Errors: {[e.msg for e in self.errors]}",
            job_id=None,
            failure_reasons=[e.msg for e in self.errors]
        )

    def status(self) -> IngestionStatus:
        """
        [ALPHA] Retrieve the status of the ingestion  from platform.

        Returns
        ----------
        IngestionStatus
            The result of the ingestion attempt

        """
        if self.raise_errors:
            raise JobFailureError(
                message=f"Ingestion creation failed: self.errors",
                job_id=None,
                failure_reasons=self.errors
            )
        else:
            return IngestionStatus.build({
                "status": IngestionStatusType.INGESTION_CREATED,
                "errors": self.errors,
            })


class IngestionCollection(Collection[Ingestion]):
    """
    [ALPHA] Represents the collection of all ingestion operations.

    Parameters
    ----------
    project_id: UUID
        Unique ID of the project this dataset collection belongs to.
    session: Session
        The Citrine session used to connect to the database.

    """

    _path_template = 'projects/{project_id}/ingestions'
    _individual_key = None
    _collection_key = None
    _resource = Ingestion

    def __init__(self, project_id: UUID, dataset_id: UUID, session: Session):
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.session = session

    def build_from_file_links(self,
                              file_links: Iterable["FileLink"],
                              *,
                              raise_errors: bool = True) -> Ingestion:
        """
        [ALPHA] Create an on-platform ingestion event based on the passed FileLink objects.

        Parameters
        ----------
        file_links: Iterable[FileLink]
            The files to ingest.
        raise_errors: bool
            Whether ingestion errors raise exceptions (vs. simply reported in the results).
            Default: True

        """
        if len(file_links) == 0:
            raise ValueError(f"No files passed.")
        invalid_links = [f for f in file_links if f.uid is None]
        if len(invalid_links) != 0:
            raise ValueError(f"{len(invalid_links)} File Links have no on-platform UID.")

        req = {
            "dataset_id": str(self.dataset_id),
            "project_id": str(self.project_id),
            "files": [
                {"dataset_file_id": str(f.uid), "file_version_uuid": str(f.version)}
                for f in file_links
            ]
        }

        try:
            response = self.session.post_resource(path=self._get_path(), json=req)
        except BadRequest as e:
            if e.api_error is not None:
                if e.api_error.validation_errors:
                    errors = [IngestionErrorTrace.from_validation_error(error)
                              for error in e.api_error.validation_errors]
                else:
                    errors = [IngestionErrorTrace(msg=e.api_error.message)]
                if raise_errors:
                    raise IngestionException(errors=errors)
                else:
                    return FailedIngestion(errors=errors)
            else:
                raise e
        return self.build({
            **response,
            "raise_errors": raise_errors
        })

    def build(self, data: dict) -> Ingestion:
        """Build an instance of an Ingestion."""
        result = Ingestion.build({**data, "session": self.session})
        return result

    def register(self, model: Ingestion) -> Ingestion:
        """Cannot register an Ingestion."""
        raise NotImplementedError("Cannot register an Ingestion.")

    def update(self, model: Ingestion) -> Ingestion:
        """Cannot update an Ingestion."""
        raise NotImplementedError("Cannot update an Ingestion.")

    def delete(self, model: Ingestion) -> Ingestion:
        """Cannot delete an Ingestion."""
        raise NotImplementedError("Cannot delete an Ingestion.")

    def list(self, *, per_page: int = 100) -> Iterator[Ingestion]:
        """Cannot list Ingestions."""
        raise NotImplementedError("Cannot list Ingestions.")



"""Record to hold branch data version update information."""
from typing import List

from citrine._rest.resource import PredictorRef, Resource
from citrine._serialization import properties as properties
from citrine._serialization.serializable import Serializable


class DataVersionUpdate(Serializable['DataVersionUpdate']):
    """Container for data updates."""

    current = properties.String('current')
    latest = properties.String('latest')

    def __init__(self,
                 *,
                 current: str,
                 latest: str):
        self.current = current
        self.latest = latest

    typ = properties.String('type', default='DataVersionUpdate')


class BranchDataUpdate(Resource['BranchDataUpdate']):
    """Branch data updates with predictors using the versions indicated."""

    data_updates = properties.List(properties.Object(DataVersionUpdate), "data_updates")
    predictors = properties.List(properties.Object(PredictorRef), "predictors")

    def __init__(self,
                 *,
                 data_updates: List[DataVersionUpdate],
                 predictors: List[PredictorRef]):
        self.data_updates = data_updates
        self.predictors = predictors


class NextBranchVersionRequest(Resource['NextBranchVersionRequest']):
    """
    Instructions for how the next version of a branch should handle its predictors.

    data_updates contains the list of data source versions to upgrade
    (current->latest), and use_predictors will wither have a <predictor_id>:latest to indicate
    the workflow should use a new version of the predictor.  Or <predictor_id>:<version #> to
    indicate that the workflow should use an existing predictor version.
    """

    data_updates = properties.List(properties.Object(DataVersionUpdate), "data_updates")
    use_predictors = properties.List(properties.Object(PredictorRef), "use_predictors")

    def __init__(self,
                 *,
                 data_updates: List[DataVersionUpdate],
                 use_predictors: List[PredictorRef]):
        self.data_updates = data_updates
        self.use_predictors = use_predictors



# flake8: noqa
from .data_source_design_space import *
from .design_space import *
from .enumerated_design_space import *
from .formulation_design_space import *
from .product_design_space import *
from .hierarchical_design_space import *



from citrine._serialization import properties
from citrine._serialization.serializable import Serializable


class SampleDesignSpaceInput(Serializable['SampleDesignSpaceInput']):
    """A Citrine Sample Design Space Execution Input.

    Parameters
    ----------
    n_candidates: int
        The number of non-correlated samples to draw from the design space domain.

    """

    n_candidates = properties.Integer("n_candidates")

    def __init__(self, *, n_candidates: int):
        self.n_candidates: int = n_candidates



from typing import Optional

from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine.informatics.constraints.constraint import Constraint

__all__ = ['ScalarRangeConstraint']


class ScalarRangeConstraint(Serializable['ScalarRangeConstraint'], Constraint):
    """Represents an inequality constraint on a real-valued material attribute.

    Parameters
    ----------
    descriptor_key: str
        the key corresponding to a descriptor
    lower_bound: float
        the minimum value in the range
    upper_bound: float
        the maximum value in the range
    lower_inclusive: bool
        if True, will include the lower bound value in the range (default: true)
    upper_inclusive: bool
        if True, will include the max value in the range (default: true)

    """

    descriptor_key = properties.String('descriptor_key')
    lower_bound = properties.Optional(properties.Float, 'min')
    upper_bound = properties.Optional(properties.Float, 'max')
    lower_inclusive = properties.Boolean('min_inclusive')
    upper_inclusive = properties.Boolean('max_inclusive')
    typ = properties.String('type', default='ScalarRange')

    def __init__(self, *,
                 descriptor_key: str,
                 lower_bound: Optional[float] = None,
                 upper_bound: Optional[float] = None,
                 lower_inclusive: Optional[bool] = None,
                 upper_inclusive: Optional[bool] = None):
        self.descriptor_key = descriptor_key

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        # we have to be careful with None and boolean values
        # None or False or True -> True, so that pattern doesn't work
        self.lower_inclusive = True
        if lower_inclusive is not None:
            self.lower_inclusive = lower_inclusive

        self.upper_inclusive = True
        if upper_inclusive is not None:
            self.upper_inclusive = upper_inclusive

    def __str__(self):
        return '<ScalarRangeConstraint {!r}>'.format(self.descriptor_key)



from typing import TypeVar

from citrine._serialization.serializable import Serializable

Self = TypeVar('Self', bound='Serializable')


class IncludeParentProperties(Serializable[Self]):
    # Calling Serializable.build() only populates the fields from the class itself and its
    # immediate parent. In order to include properties from the grandparent, the parent must be
    # deserialized, then merged with the results of the child's deserialization.
    @classmethod
    def build_with_parent(cls, data: dict, base_cls) -> Self:
        resource = super().build(data)

        from citrine._serialization import properties
        metadata_properties = properties.Object(base_cls).deserialize(data)

        resource.__dict__.update(metadata_properties.__dict__)
        return resource

    # This is required to allow deserialization to create an instance to read the properties.
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)



from typing import List, Union, Optional
from uuid import UUID

from citrine._rest.engine_resource import ModuleEngineResource
from citrine._serialization import properties
from citrine.informatics.design_spaces.design_space import DesignSpace
from citrine.informatics.dimensions import Dimension

__all__ = ['ProductDesignSpace']


class ProductDesignSpace(ModuleEngineResource['ProductDesignSpace'], DesignSpace):
    """A Cartesian product of design spaces.

    Factors can be other design spaces and/or univariate dimensions.

    Parameters
    ----------
    name:str
        the name of the design space
    description:str
        the description of the design space
    subspaces: List[Union[UUID, DesignSpace]]
        the list of subspaces to combine, either design spaces defined in-line
        or UUIDs that reference design spaces on the platform
    dimensions: list[Dimension]
        univariate dimensions that are factors of the design space; can be enumerated or continuous

    """

    subspaces = properties.List(properties.Object(DesignSpace), 'data.instance.subspaces',
                                default=[])
    dimensions = properties.Optional(
        properties.List(properties.Object(Dimension)), 'data.instance.dimensions'
    )

    typ = properties.String('data.instance.type', default='ProductDesignSpace',
                            deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 subspaces: Optional[List[Union[UUID, DesignSpace]]] = None,
                 dimensions: Optional[List[Dimension]] = None):
        self.name: str = name
        self.description: str = description
        self.subspaces: List[Union[UUID, DesignSpace]] = subspaces or []
        self.dimensions: List[Dimension] = dimensions or []

    def _post_dump(self, data: dict) -> dict:
        data = super()._post_dump(data)
        for i, subspace in enumerate(data['instance']['subspaces']):
            if isinstance(subspace, dict):
                # embedded design spaces are not modules, so only serialize their config
                data['instance']['subspaces'][i] = subspace['instance']
        return data

    @classmethod
    def _pre_build(cls, data: dict) -> dict:
        for i, subspace_data in enumerate(data['data']['instance']['subspaces']):
            if isinstance(subspace_data, dict):
                data['data']['instance']['subspaces'][i] = DesignSpace.wrap_instance(subspace_data)
        return data

    def __str__(self):
        return '<ProductDesignSpace {!r}>'.format(self.name)



# flake8: noqa
from .categorical_constraint import *
from .constraint import *
from .ingredient_count_constraint import *
from .ingredient_fraction_constraint import *
from .label_fraction_constraint import *
from .scalar_range_constraint import *
from .integer_range_constraint import *
from .ingredient_ratio_constraint import *



"""Top-level class for all object template objects and collections thereof."""
from abc import ABC
from typing import TypeVar

from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import String
from citrine.resources.templates import Template, TemplateCollection
from gemd.entity.template.base_template import BaseTemplate as GEMDTemplate


class ObjectTemplate(Template, GEMDTemplate, ABC):
    """
    An abstract object template object.

    ObjectTemplate must be extended along with `Resource`
    """

    name = String('name')
    description = PropertyOptional(String(), 'description')


ObjectTemplateResourceType = TypeVar("ObjectTemplateResourceType", bound="ObjectTemplate")


class ObjectTemplateCollection(TemplateCollection[ObjectTemplateResourceType], ABC):
    """A collection of one kind of object template object."""



"""Top-level class for all object run objects and collections thereof."""
from abc import ABC
from typing import TypeVar

from citrine.resources.data_objects import DataObject, DataObjectCollection
from gemd.entity.object.has_spec import HasSpec


class ObjectRun(DataObject, HasSpec, ABC):
    """
    An abstract object run object.

    ObjectRun must be extended along with `Resource`
    """


ObjectRunResourceType = TypeVar("ObjectRunResourceType", bound="ObjectRun")


class ObjectRunCollection(DataObjectCollection[ObjectRunResourceType], ABC):
    """A collection of one kind of object run object."""



from typing import List, Union
from uuid import UUID

from deprecation import deprecated

from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.informatics.data_sources import DataSource
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.predictors import PredictorNode, GraphPredictor


# Not a full Collection since CRUD operations are not valid for Descriptors
class DescriptorMethods:
    """A set of methods returning descriptors which require connection to the backend."""

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session

    def from_predictor_responses(self, *, predictor: Union[GraphPredictor, PredictorNode],
                                 inputs: List[Descriptor]) -> List[Descriptor]:
        """
        Get responses for a predictor, given an input space.

        Parameters
        ----------
        predictor : Union[Predictor, PredictorNode]
            Either a single predictor node or full predictor
             whose available responses are to be computed.
        inputs : List[Descriptor]
            The input space to the predictor.

        Returns
        -------
        List[Descriptor]
            The computable responses of the predictor given the provided input space (as
            descriptors).

        """
        if isinstance(predictor, GraphPredictor):
            predictor_data = predictor.dump()["instance"]
        else:
            predictor_data = predictor.dump()

        response = self.session.post_resource(
            path=format_escaped_url('/projects/{}/material-descriptors/predictor-responses',
                                    self.project_id),
            json={
                'predictor': predictor_data,
                'inputs': [i.dump() for i in inputs]
            }
        )
        return [Descriptor.build(r) for r in response['responses']]

    def from_data_source(self, *, data_source: DataSource) -> List[Descriptor]:
        """
        Get all descriptors associated with a data source.

        Parameters
        ----------
        data_source : DataSource
            A CSVDataSource or AraTableDataSource to get descriptors for.

        Returns
        -------
        List[Descriptor]
            The list of descriptors associated with the given data_source.

        """
        response = self.session.post_resource(
            path=format_escaped_url('/projects/{}/material-descriptors/from-data-source',
                                    self.project_id),
            json={
                'data_source': data_source.dump()
            }
        )
        return [Descriptor.build(r) for r in response['descriptors']]

    @deprecated(deprecated_in="2.31.0", removed_in="3.0.0",
                details="Use from_data_source instead.")
    def descriptors_from_data_source(self, *, data_source: DataSource) -> List[Descriptor]:
        """[DEPRECATED] See from_data_source."""
        return self.from_data_source(data_source=data_source)



import citrine.jobs.job
from citrine.jobs.job import _poll_for_job_completion  # noqa
from citrine._utils.functions import MigratedClassMeta


class JobSubmissionResponse(citrine.jobs.job.JobSubmissionResponse,
                            deprecated_in="2.22.1",
                            removed_in="3.0.0",
                            metaclass=MigratedClassMeta):
    """
    A response to a submit-job request for the job submission framework.

    This is returned as a successful response from the remote service.

    Importing from this package is deprecated; import from citrine.jobs.job instead.

    """


class TaskNode(citrine.jobs.job.TaskNode,
               deprecated_in="2.22.1",
               removed_in="3.0.0",
               metaclass=MigratedClassMeta):
    """
    Individual task status.

    The TaskNode describes a component of an overall job.

    Importing from this package is deprecated; import from citrine.jobs.job instead.

    """


class JobStatusResponse(citrine.jobs.job.JobStatusResponse,
                        deprecated_in="2.22.1",
                        removed_in="3.0.0",
                        metaclass=MigratedClassMeta):
    """
    A response to a job status check.

    The JobStatusResponse summarizes the status for the entire job.

    Importing from this package is deprecated; import from citrine.jobs.job instead.

    """



"""Resources that represent both individual and collections of sample design space executions."""
from typing import Union, Iterator
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._session import Session
from citrine.informatics.executions.sample_design_space_execution import SampleDesignSpaceExecution
from citrine.informatics.design_spaces.sample_design_space import SampleDesignSpaceInput
from citrine.resources.response import Response


class SampleDesignSpaceExecutionCollection(Collection["SampleDesignSpaceExecution"]):
    """A collection of SampleDesignSpaceExecutions."""

    _api_version = 'v3'
    _path_template = '/projects/{project_id}/design-spaces/{design_space_id}/sample'
    _individual_key = None
    _collection_key = 'response'
    _resource = SampleDesignSpaceExecution

    def __init__(self, project_id: UUID, design_space_id: UUID, session: Session):
        self.project_id: UUID = project_id
        self.design_space_id: UUID = design_space_id
        self.session: Session = session

    def build(self, data: dict) -> SampleDesignSpaceExecution:
        """Build an individual SampleDesignSpaceExecution."""
        execution = SampleDesignSpaceExecution.build(data)
        execution._session = self.session
        execution.project_id = self.project_id
        execution.design_space_id = self.design_space_id
        return execution

    def trigger(
        self, sample_design_space_input: SampleDesignSpaceInput
    ) -> SampleDesignSpaceExecution:
        """Trigger a sample design space execution."""
        path = self._get_path()
        request_dict = sample_design_space_input.dump()
        data = self.session.post_resource(path, request_dict, version=self._api_version)
        return self.build(data)

    def register(self, model: SampleDesignSpaceExecution) -> SampleDesignSpaceExecution:
        """Cannot register an execution."""
        raise NotImplementedError("Cannot register a SampleDesignSpaceExecution.")

    def update(self, model: SampleDesignSpaceExecution) -> SampleDesignSpaceExecution:
        """Cannot update an execution."""
        raise NotImplementedError("Cannot update a SampleDesignSpaceExecution.")

    def list(self, *,
             per_page: int = 10,
             ) -> Iterator[SampleDesignSpaceExecution]:
        """
        Paginate over the elements of the collection.

        Leaving page and per_page as default values will yield all elements in the
        collection, paginating over all available pages.

        Parameters
        ---------
        page: int, optional
            The "page" of results to list. Default is to read all pages and yield
            all results.  This option is deprecated.
        per_page: int, optional
            Max number of results to return per page. Default is 100.  This parameter
            is used when making requests to the backend service.  If the page parameter
            is specified it limits the maximum number of elements in the response.

        Returns
        -------
        Iterator[ResourceType]
            Resources in this collection.

        """
        return self._paginator.paginate(page_fetcher=self._fetch_page,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Sample Design Space Executions cannot be deleted or archived."""
        raise NotImplementedError(
            "Sample Design Space Executions cannot be deleted"
        )



import functools
from typing import Iterator, Optional, Union
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine.exceptions import NotFound
from citrine.resources.data_version_update import BranchDataUpdate, NextBranchVersionRequest
from citrine.resources.design_workflow import DesignWorkflowCollection
from citrine.resources.experiment_datasource import (ExperimentDataSourceCollection,
                                                     ExperimentDataSource)


class Branch(Resource['Branch']):
    """
    A project branch.

    A branch is a container for design workflows.
    """

    name = properties.String('data.name')
    uid = properties.Optional(properties.UUID(), 'id')
    archived = properties.Boolean('metadata.archived', serializable=False)
    created_at = properties.Optional(properties.Datetime(), 'metadata.created.time',
                                     serializable=False)
    updated_at = properties.Optional(properties.Datetime(), 'metadata.updated.time',
                                     serializable=False)
    # added in v2
    root_id = properties.UUID('metadata.root_id', serializable=False)
    version = properties.Integer('metadata.version', serializable=False)

    project_id: Optional[UUID] = None

    def __init__(self,
                 name: str,
                 *,
                 session: Optional[Session] = None):
        self.name: str = name
        self.session: Session = session

    def __str__(self):
        return f'<Branch {self.name!r}>'

    @property
    def design_workflows(self) -> DesignWorkflowCollection:
        """Return a resource representing all workflows contained within this branch."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot initialize workflow without project reference!')
        return DesignWorkflowCollection(project_id=self.project_id,
                                        session=self.session,
                                        branch_id=self.uid)

    @property
    def experiment_datasource(self) -> Optional[ExperimentDataSource]:
        """Return this branch's experiment data source, or None if one doesn't exist."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot retrieve datasource without project reference!')
        erds = ExperimentDataSourceCollection(project_id=self.project_id, session=self.session)
        branch_erds_iter = erds.list(branch_id=self.uid, version="latest")
        return next(branch_erds_iter, None)

    def _post_dump(self, data: dict) -> dict:
        # Only the data portion of an entity is sent to the server.
        data = data["data"]
        return data


class BranchCollection(Collection[Branch]):
    """A collection of Branches."""

    _path_template = '/projects/{project_id}/branches'
    _individual_key = None
    _collection_key = 'response'
    _resource = Branch
    _api_version = 'v2'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id: UUID = project_id
        self.session: Session = session

    def build(self, data: dict) -> Branch:
        """
        Build an individual branch from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the branch.

        Return
        -------
        Branch
            The branch created from data.

        """
        branch = Branch.build(data)
        branch.session = self.session
        branch.project_id = self.project_id
        return branch

    def get_by_root_id(self, *, branch_root_id: Union[UUID, str]) -> Branch:
        """
        Given a branch root ID, retrieve the latest version of the branch.

        Parameters
        ---------
        branch_root_id:  Union[UUID, str]
            Unique identifier of the branch root

        Returns
        -------
        Branch
            The latest version of the branch.

        """
        params = {"root": str(branch_root_id), "version": "latest"}
        branch = next(self._list_with_params(per_page=1, **params), None)
        if branch:
            return branch
        else:
            raise NotFound.build(
                message=f"Branch root '{branch_root_id}' not found",
                method="GET",
                path=self._get_path(),
                params=params
            )

    def list(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List active branches using pagination.

        Yields all active branches, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            Active branches in this collection.

        """
        return super().list(per_page=per_page)

    def list_archived(self, *, per_page: int = 20) -> Iterator[Branch]:
        """
        List archived branches using pagination.

        Yields all archived branches, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of branches to return per page. Default is 20. This parameter is used when
            making requests to the backend service.

        Returns
        -------
        Iterator[Branch]
            Archived branches in this collection.

        """
        return self._list_with_params(per_page=per_page, archived=True)

    def _list_with_params(self, *, per_page, **kwargs):
        fetcher = functools.partial(self._fetch_page, additional_params=kwargs)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def archive(self, uid: Union[UUID, str] = None):
        """
        Archive a branch.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the branch to archive

        """
        url = self._get_path(uid, action="archive")
        data = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(data)

    def restore(self, uid: Union[UUID, str] = None):
        """
        Restore an archived branch.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the branch to restore

        """
        url = self._get_path(uid, action="restore")
        data = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(data)

    def update_data(self,
                    branch: Union[UUID, str, Branch],
                    *,
                    use_existing: bool = True,
                    retrain_models: bool = False) -> Optional[Branch]:
        """
        Automatically advance the branch to the next version.

        If there are no newer versions of data sources used by this branch this method returns
        without doing anything

        Parameters
        ----------
        branch: Union[UUID, str, Branch]
            Branch Identifier or Branch object

        use_existing: bool
            If true the workflows in this branch will use existing predictors that are using
            the latest versions of the data sources and are ready to use.

        retrain_models: bool
            If true, when new versions of models are created, they are automatically
            scheduled for training.

        Returns
        -------
        Branch
            The new branch record after version update or None if no update

        """
        if not isinstance(branch, Branch):
            branch = self.get(branch)
        version_updates = self.data_updates(branch.uid)
        # If no new data sources, then exit, nothing to do
        if len(version_updates.data_updates) == 0:
            return None

        use_predictors = []
        if use_existing:
            use_predictors = version_updates.predictors

        branch_instructions = NextBranchVersionRequest(data_updates=version_updates.data_updates,
                                                       use_predictors=use_predictors)
        branch = self.next_version(branch.root_id,
                                   branch_instructions=branch_instructions,
                                   retrain_models=retrain_models)
        return branch

    def data_updates(self, uid: Union[UUID, str]) -> BranchDataUpdate:
        """
        Get data updates for a branch.

        Parameters
        ----------
        uid:  Union[UUID, str]
            Unique identifier of the branch to get data updates for

        Returns
        -------
        BranchDataUpdate
            A list of data updates and compatible predictors

        """
        path = self._get_path(uid=uid, action="data-version-updates-predictor")
        data = self.session.get_resource(path, version=self._api_version)
        return BranchDataUpdate.build(data)

    def next_version(self,
                     branch_root_id: Union[UUID, str],
                     *,
                     branch_instructions: NextBranchVersionRequest,
                     retrain_models: bool = True):
        """
        Move a branch to the next version.

        Parameters
        ----------
        branch_root_id:  Union[UUID, str]
            Unique identifier of the branch root to advance to next version

        branch_instructions: NextBranchVersionRequest
            Instructions for how the next version of a branch should handle its predictors
            when the workflows are cloned.  data_updates contains the list of data source
            versions to upgrade (current->latest), and use_predictors will either have a
            <predictor_id>:latest to indicate the workflow should use a new version of
            the predictor.  Or <predictor_id>:<version #> to indicate that the workflow
            should use an existing predictor version.

        retrain_models: bool
            If true, when new versions of models are created, they are automatically
            scheduled for training

        Returns
        -------
        Branch
            The new branch record after version update

        """
        path = self._get_path(action="next-version-predictor")
        data = self.session.post_resource(path, branch_instructions.dump(),
                                          version=self._api_version,
                                          params={
                                              'root': str(branch_root_id),
                                              'retrain_models': retrain_models})
        return self.build(data)



import time
from pprint import pprint
from typing import Union

from citrine._rest.collection import Collection
from citrine._rest.asynchronous_object import AsynchronousObject
from citrine.informatics.executions.design_execution import DesignExecution
from citrine.informatics.executions.generative_design_execution import GenerativeDesignExecution
from citrine.informatics.executions.sample_design_space_execution import SampleDesignSpaceExecution
from citrine.informatics.executions import PredictorEvaluationExecution
from citrine.informatics.modules import Module


class ConditionTimeoutError(RuntimeError):
    """Error that is raised when timeout is reached but the checked condition is still False."""

    pass


def _print_string_status(
    status: str, start_time: float, line_start: str = "", line_end: str = "\r"
):
    print(
        "{}Status = {:<25}Elapsed time".format(line_start, status),
        " = {}s".format(str(int(time.time() - start_time)).rjust(3)),
        end=line_end,
    )


def wait_for_asynchronous_object(
    *,
    resource: AsynchronousObject,
    collection: Collection[AsynchronousObject],
    print_status_info: bool = False,
    timeout: float = 1800.0,
    interval: float = 3.0
) -> AsynchronousObject:
    """
    Wait until an asynchronous object has finished.

    This could be a module, workflow, workflow execution, or report.

    Parameters
    ----------
    resource: AsynchronousObject
        a module, workflow, workflow execution, or report to monitor
    collection: Collection[AsynchronousObject]
        Collection containing the resource
    print_status_info: bool
        Whether to print status info, by default False
    timeout : float
        Maximum time spent waiting, in seconds, by default 1800.0
    interval: float
        Inquiry interval in seconds, by default 3.0

    Returns
    -------
    AsynchronousObject
        The resource after the asynchronous work has finished or timed out.

    Raises
    ------
    ConditionTimeoutError
        If fails to finish within timeout

    """
    start = time.time()

    def is_finished():
        current_resource = collection.get(resource.uid)
        if print_status_info:
            _print_string_status(current_resource.status, start)
        return not current_resource.in_progress()

    while not is_finished() and (time.time() - start < timeout):
        time.sleep(interval)
    if not is_finished():
        raise ConditionTimeoutError(
            "Timeout of {timeout_length} seconds "
            "reached, but task {uid} is still in progress".format(
                timeout_length=timeout, uid=resource.uid)
        )

    current_resource = collection.get(resource.uid)
    if print_status_info and hasattr(current_resource, 'status_detail'):
        print("\nStatus info:")
        pprint([detail.msg for detail in current_resource.status_detail])
    return current_resource


def wait_while_validating(
    *,
    collection: Collection[Module],
    module: Module,
    print_status_info: bool = False,
    timeout: float = 1800.0,
    interval: float = 3.0,
) -> Module:
    """
    Wait until module is validated.

    Parameters
    ----------
    collection : Collection[Module]
        Collection containing module
    module : Module
        Module in Collection
    print_status_info : bool, optional
        Whether to print status info, by default False
    timeout : float, optional
        Maximum time spent inquiring in seconds, by default 1800.0
    interval : float, optional
        Inquiry interval in seconds, by default 3.0

    Returns
    -------
    Module
        Module in Collection after validation

    Raises
    ------
    ConditionTimeoutError
        If fails to validate within timeout

    """
    return wait_for_asynchronous_object(resource=module, collection=collection,
                                        print_status_info=print_status_info, timeout=timeout,
                                        interval=interval)


def wait_while_executing(
    *,
    collection: Union[
        Collection[PredictorEvaluationExecution],
        Collection[DesignExecution],
        Collection[GenerativeDesignExecution],
        Collection[SampleDesignSpaceExecution]
    ],
    execution: Union[
        PredictorEvaluationExecution,
        DesignExecution,
        GenerativeDesignExecution,
        SampleDesignSpaceExecution
    ],
    print_status_info: bool = False,
    timeout: float = 1800.0,
    interval: float = 3.0,
) -> Union[
        PredictorEvaluationExecution,
        DesignExecution,
        GenerativeDesignExecution,
        SampleDesignSpaceExecution,
]:
    """
    Wait until execution is finished.

    Parameters
    ----------
    execution : Union[PredictorEvaluationExecution, DesignExecution]
        an execution to monitor
    print_status_info : bool, optional
        Whether to print status info, by default False
    timeout : float, optional
        Maximum time spent inquiring in seconds, by default 1800.0
    interval : float, optional
        Inquiry interval in seconds, by default 3.0
    collection : Union[Collection[PredictorEvaluationExecution], Collection[DesignExecution]]
        Collection containing executions

    Returns
    -------
    Union[PredictorEvaluationExecution, DesignExecution]
        the updated execution after it has finished executing


    Raises
    ------
    ConditionTimeoutError
        If fails to finish execution within timeout

    """
    return wait_for_asynchronous_object(resource=execution, collection=collection,
                                        print_status_info=print_status_info, timeout=timeout,
                                        interval=interval)



from typing import Optional

from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine.informatics.constraints.constraint import Constraint

__all__ = ['IntegerRangeConstraint']


class IntegerRangeConstraint(Serializable['IntegerRangeConstraint'], Constraint):
    """[ALPHA] Represents an inequality constraint on an integer-valued material attribute.

    Warning: IntegerRangeConstraints are not fully supported by the Citrine Platform web interface
    and may cause unexpected issues until resolved.

    Parameters
    ----------
    descriptor_key: str
        the key corresponding to a descriptor
    lower_bound: int
        the minimum value in the range
    upper_bound: int
        the maximum value in the range
    lower_inclusive: bool
        if True, will include the lower bound value in the range (default: true)
    upper_inclusive: bool
        if True, will include the max value in the range (default: true)

    """

    descriptor_key = properties.String('descriptor_key')
    lower_bound = properties.Optional(properties.Float, 'min')
    upper_bound = properties.Optional(properties.Float, 'max')
    typ = properties.String('type', default='IntegerRange')

    def __init__(self, *,
                 descriptor_key: str,
                 lower_bound: Optional[int] = None,
                 upper_bound: Optional[int] = None):
        self.descriptor_key = descriptor_key
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __str__(self):
        return '<IntegerRangeConstraint {!r}>'.format(self.descriptor_key)
